\ words for catching interrupts.  To use, define a word which does
\ what you want to do when an interrupt comes along.  Assign it to
\ doint as follows:   ' yourword is doint
\ Let's assume that you want to catch level 3 interrupts.
\ Type    3 connect-int   to replace the interrupt vector
\ Then    2 spl           to set the CPU priority low enough to respond
\ to level 3 interrupts.  You can use any level that you want to, but
\ level 7 will interfere with the monitor
\ In general, it is not possible to print something during the interrupt
\ service routine, then return from interrupt.  I think this is because
\ the monitor is not completely reentrant, so if the interrupt happens
\ to occur while the monitor is executing, and you then try to print
\ something, the monitor is reentered and gets confused.
\ If your interrupt service word prints out anything, it is necessary
\ to do a "quit" so that control goes to forth and the return-from-interrupt
\ is not done.

hex

code retint ( return from interrupt )
  long sp )+ ffff # movem   ( restore registers )
  rte
c;

: print-interrupt ." Interrupt! " quit ;
defer doint
' print-interrupt is doint

: intser doint retint ;

code takeint ( low-level interrupt handler )
  long ffff # sp -) movem   ( save registers )
( assumes that the monitor doesn't mess with a3.  If this is not the case )
( it may be necessary to do a full-blown context-switch )
           a3 d7    lmove   ( fix up relocation pointer )
  ' intser >body n# d7 rmove  ( get and relocate addr. of service routine )
  d7 ip lmove               ( go to it )
c;

: connect-int ( vector# -- )
  ['] takeint >body rel ( interrupt-level d.takeint-address )
  rot /l * n->l l!l
;
decimal
