\ Words to allocate/free memory.  These words are implementation-dependent.
\ Memory is allocated from the space between the top of the dictionary
\ and the parameter stack.   The first memory to be allocated is just below
\ the parameter stack area, and succeding allocations proceed downward from
\ there.

\ Implementation notes:
\ N.B. this is not a general-purpose allocator -- only the last memory
\ allocated may be freed.  It is correct to view the memory space that
\ is allocated as a stack.  Allocating memory adds nbytes to that stack.
\ Freeing it pops nbytes from that stack.
\ The address argument to free-mem is provided so that a general-purpose
\ allocator may be substitutedsomeday.

create alloc-mem.f
hex
200 constant ps-size
variable next-free-mem  0 next-free-mem !
: init-malloc  \ Doesn't do anything if already initialized
  next-free-mem @ 0=
  if  sp0 @  ps-size -   next-free-mem !  then
;
: alloc-mem ( nbytes -- address )
  next-free-mem @ swap -  dup next-free-mem !
;
: free-mem ( address nbytes -- )  ( address is not used here )
  next-free-mem +!
  drop    ( address not used in this implementation )
;
decimal
