.nr PS 9
.nr ps 9
.UH H
Command Line Editing with Command Completion
.UH H
Mitch Bradley
.SH Abstract
Given the popularity of screen editors and the `what you see is what you get'
principle, it is surprising that most systems do not provide similar
capabilities for editing command lines.
.PP
This paper presents a screen-oriented command line editor, including a
command completion feature.  Command completion allows a user to abbreviate
any name by typing the first few letters.  The computer then completes the
name if it is possible to uniquely determine which name was meant.
.SH
Introduction
.PP
Given a choice of editors, most people will choose a screen-oriented
editor over a line editor.  The `what you see is what you get' principle
seems well suited to human capabilities.  Most computer systems provide
a full-featured screen editor for text files, but few provide the same
capabilities for editing lines that are being typed to the command
interpreter.  Since a lot of interaction with a computer system
is done by typing command lines, it seems sensible to provide powerful
editing capabilities for composing those command lines.  Nevertheless, most
systems provide only very-limited commmand line editing, usually
`erase previous character', `erase entire line', and sometimes
`erase previous word'.
.PP
In FORTH, the equivalent of typing a command line is typing a line to the
text interpreter, usually under control of the word EXPECT.  I present a
full-featured screen-oriented intra-line editor to replace EXPECT.
The command-set includes, among others, 
`forward/backward character/word', `erase previous/next character/word',
and `to beginning/end of line'.  The fanciest feature of this editor is
called `command completion'.
.PP
Command completion is a technique for automatically abbreviating word names.
With command completion, a user need only type the first few characters of a
name, and the system will fill in the rest.  Only enough characters to
uniquely identify the word from among the possible choices are required.
If the user doesn't type enough characters to uniquely identify the word,
the system fills out the word as far as possible, beeping to indicate that
more characters are needed to select a unique word.  The user may ask at any
time to see all the current choices.  For instance, if the dictionary
contains two words `expand-word' and `expand-consciousness', the user could
type `ex' and the command completion command character.  The system would
expand the word on the display to `expand-' and then beep.  The user could
then add `w' and again type the command completion command character, at
which time the system would complete the word to `expand-word '.
.PP
Command completion has been around for some time.  The TENEX\**
.FS
`TENEX' is a registered trademark of Bolt, Beranek, and Newman, Inc.
.FE
operating system [BOBR] for the DEC-10\**
.FS
`DEC-10' is a trademark of Digital Equipment Corporation.
.FE
family of computers has it, as do TOPS-20\**,
.FS
`TOPS-20' is a trademark of Digital Equipment Corporation.
.FE
some versions of Emacs, several of Control Data Corporation's products,
and even the microcomputer file transfer program Kermit.
It tends to be a very popular feature; people miss it when
they move from a system with it to one without.  Furthermore, command
completion removes the major objection to using long, readable command names,
because it is no longer painful to type them.
.SH
Editor Commands
.PP
The intra-line editor is based on Emacs [STAL].  Emacs is an extensible
screen-oriented editor that was developed about 10 years ago at the MIT AI
lab by Richard Stallman.  Since then, Emacs or something like it has been
implemented on lots of different machines.  A few microcomputer editors are
based on Emacs, for example MINCE (which stands for Mince Is Not Complete
Emacs).  Emacs tends to inspire fanatical devotion among its fans in the
same way that FORTH does.  I freely admit to being an Emacs fan myself.
(I usually run FORTH as a subprocess INSIDE Emacs!).
.PP
Emacs editing commands are invoked by typing control characters.  When a
printable character is typed, it is simply added to the text being edited.
When a control character is typed, a table is consulted to determine which
editing command to perform.  The association between a particular control
character and a particular editing function may be easily changed by the
user at any time.
.PP
There is a philosophical debate about the relative merits of editors which
rely on control characters versus those which use printable characters as
commands and have a separate `insert mode' for typing text.  This editor
uses control characters.  If you prefer the other style, it should be easy
to modify to suit your taste.
.PP
The editor always displays the line being edited on the screen.  The
cursor may be anywhere within the line.  The character immediately to the
left of the cursor is the `previous' character.  The character underneath
the cursor is the `next' character.  If a printable character is typed, it is
inserted between the previous and next characters, and the rest of the line
is shifted to the right to make room.  It is useful to think of the cursor
as being between the previous and next characters, even though it is
displayed on top of the next character.  A `word' is a sequence of non-blank
characters, just like in FORTH.
.PP
.B forward-character
moves the cursor forward one character, unless it's already at the end of
the line.
.PP
.B backward-character
moves the cursor backward one character, unless it's already at the beginning
of the line.
.PP
.B forward-word
moves the cursor forward one word, unless it's already at the end of
the line.  The cursor is left so that the last character in the word is the
`previous' character.  If the cursor starts out in the middle of a word, it
is moved to the end of that word.  Otherwise it skips blanks until a word is
found, and moves to the end of that word.
.PP
.B backward-word
moves the cursor backward one word, unless it's already at the beginning of
the line.  The cursor is left so that the first character in the word is the
`next' character.  If the cursor starts out in the middle of a word, it
is moved to the beginning of that word.  Otherwise it skips blanks backward
until a word is found, and moves to the beginning of that word.
.PP
.B end-of-line
moves the cursor to the end of the line.
.PP
.B beginning-of-line
moves the cursor to the beginning of the line.
.PP
.B erase-next-character
erases the next character (the one under the cursor), unless the cursor is
already at the end of the line, in which case it does nothing.  The rest of
the line moves to the left to fill in the vacated position.
.PP
.B erase-previous-character
erases the previous character (the one under the cursor), unless the cursor is
already at the beginning of the line, in which case it does nothing.  The
rest of the line moves to the left to fill in the vacated position.
.PP
.B erase-next-word
erases the next word, unless the cursor is
already at the end of the line, in which case it does nothing.
If the cursor starts in the middle of a word, only the end of the word is
erased; the portion to the left of the cursor remains intact.  If the cursor
starts in the middle of a series of blanks, everything up to the end of the
next word is erased.
.PP
.B erase-previous-word
erases the previous word, unless the cursor is
already at the beginning of the line, in which case it does nothing.
If the cursor starts in the middle of a word, only the beginning of
the word is
erased; the portion to the right of the cursor remains intact.  If the cursor
starts in the middle of a series of blanks, everything back to the beginning
of the previous word is erased.
.PP
.B kill-to-end-of-line
erases everything to the right of the cursor, including the character under
the cursor.
.PP
.B insert-character
inserts the last character typed between the previous and next characters.
The rest of the line is moved to the right to make room for the new
character.  This is the function that is executed when a printable character
is typed.
.PP
.B quote-next-character
gets a new character from the keyboard, and inserts it into the line between
the previous and next characters, even if the character is a control
character.  This is the way to force a control character into the line,
instead of the control character causing some editing function to be
performed.
.PP
.B finish-line
terminates the entry of the line.  EXPECT then returns the line to its
caller.  This function is normally executed when carriage return is typed.
.PP
.B retype-line
redisplays the line on the screen, in case something has happened
to mess up the display.
.PP
.B grab-last-line
retrieves the previously-typed line so that it may be edited.  This is
useful for making minor modifications to the previous line.
.PP
.B expand-word
is the command completion function.  The word underneath the cursor is used
as a search string.  The dictionary is searched according to the CONTEXT
search order for words whose first characters match the search string.  If
there is only one match, the word on the screen is filled out to the matched
word and a blank is added to the end.  If more than one match is found, the
word is filled out as far as all the matches agree, and the terminal 
bell is rung.  If no matches are found, characters are successively removed
from the end of the search string until at least one match is found.  The
word on the screen is truncated to the length of the new wearch string, and
the bell is rung.
.PP
.B show-matches
displays a list of all those words which match the word underneath the
cursor, according to the search rules described in expand-word.
.SH
Default key assignments
.PP
Here is the code that assigns keys to functions.  ^x means control-x, i.e.
hold down the control key and type x.  ^[x means escape-x, i.e. type the
escape key then type x\**.
.FS
In Emacs parlance, escape-x is sometimes called `meta-x',
because there were some funny keyboards at MIT and Stanford that
had several shift keys.
One shift key was called `meta', and was used in place of escape.
One style of keyboard had about five shift keys, and you had to type
chords to get certain things done.
.FE
.DS
: ^F forward-character ;
: ^B backward-character ;
: ^A beginning-of-line ;
: ^E end-of-line ;
: ^D erase-next-character ;
: ^H erase-previous-character ; ( backspace )
: ^K kill-to-end-of-line ;
: ^M finish-line ;  ( carriage return )
: ^Q quote-next-character ;
: ^L retype-line ;
: ^[h erase-previous-word ;
: ^[d erase-next-word ;
: ^[f forward-word ;
: ^[b backward-word ;
: ^[= grab-last-line ;
: ^J expand-word ;
: ^_ do-show ;
.DE
.SH
Implementation
.PP
There are a couple of `tried-and-true' methods that have been frequently
used for implementing control-character-driven screen editors.  One is to use
a case statement or a nested conditional to decide what to do with each
control character.  Another method is to use a jump table or execution vector
table that is indexed by the numerical value of the control character.
The first is difficult to modify without recompilation, which is especially
painful if the function is deep in the FORTH kernel as EXPECT usually is.
The second is easy to modify, but special words are required to conveniently
see what functions are currently associated with what keys.
.PP
The technique used here follows from the observation that the FORTH text
interpreter is equivalent to a giant extensible CASE statement, whose CASEs
are the words in the dictionary and whose search key is the word being
interpreted.  When a control key is typed, this editor converts the key to a
text string of the form ^X (for control-X), and tries to interpret a word with
that name from a special vocabulary.  If no such word is found, the editor
just beeps instead of ABORT'ing.
.PP
This has several nice properties. First, it is very easy to associate a
function with a particular control key.  For instance, suppose that you want
to execute `expand-word' when control-j is typed.  Make the control
character vocabulary the CURRENT one, and do:
.PP
: ^J expand-word ;
.PP
The current set of associations between control characters and functions may
be easily seen using existing tools.  WORDS (the F83 name for VLIST) will
show you the
control characters that are defined, and the decompiler will show you the
function executed by a particular control character.
.PP
Keyboard macros are easy to implement using this scheme.  Just define the
behavior of any control character you wish with a colon definition.  The
definition can either be in terms of the function names or of
previously-defined control characters.
.PP
Multiple-keystroke commands are also easy.  For instance, suppose you want to
use escape-x as a command, where escape-x means type the escape key, then
type the x key.
Since escape is the same as control-[, just define ^[ to get a keyboard
character, append it to the control-character search string, and search the
vocabulary again.  Any two-charcter escape sequence may then be defined in
the obvious way, e.g.
.PP
: ^[x ( whatever you want escape-x to do ) ;
.PP
It may be argued that this technique is slow.  However, compared to the
speed at which people can type, it is plenty fast.  120 words-per-minute
corresponds to 10 characters-per-second, which is 100 msec-per-character.
This is a long time for a computer, and hardly anybody can even approach
120 words-per-minute.  Besides, the whole dictionary doesn't have to be
searched, only the control character vocabulary (presupposing sealed
vocabularies).
.PP
Another possible objection is that it may seem difficult to implement.
It only took me a few minutes;  you may look at the code and judge
for yourself about the complexity.
.PP
The command completion code should be relatively easy to follow.
The algorithm is pretty simple:  given a search string, go through the
dictionary and make a list of the NFA's of all the words whose initial
substrings match the search string.  If the list contains one match, all is
well; expand the word to be the same as the matched word.  If the list
contains several words, find the longest initial substring common to all the
matches, and expand the word to that substring.  If the list contains no
matches, remove the last character from the search string and start over.
Most of the hard work is involved in getting the substring match code right,
considering that the headers in the dictionary headers may have tag bits in
the character strings.
.SH
FORTH dialect
.PP
This code was written using a version of the public domain F83 system by
Perry and Laxen.  The system has been modified somewhat to run on a 68000
Unix\**
.FS
Unix is a trademark of Bell Laboratories.
.FE
system (actually a Sun Workstation\**
.FS
Sun Workstation is a trademark of Sun Microsystems, Inc.
.FE
).  Several implementation
dependencies exist.  In many cases these were unavoidable, and point out
deficiencies in the standard.  In particular:  It is assumed that EXPECT is
vectored so that it may be replaced with the new version.  If your EXPECT
isn't vectored, you have to find a way to patch the new one in.
There is a word `vfind' which takes as arguments the address of a string to
look for and the cfa of a vocabulary.  It then searches just that vocabulary.
This word is easy to implement on any particular system with sealed
vocabularies.  It may in fact be implementable in 83-standard FORTH,
but saving and restoring CONTEXT is painful, so I implemented it in
the obvious way for my system.  A well-factored kernel ought to have
such a word present.
.PP
The other implementation dependency is in the command
completion package.  So far as I know, the 83 standard does not provide anyy
standard way of looking through the dictionary other than with FIND.  FIND
can't cope with looking for matching initial substrings, so there is a word
`find-candidates' which looks through the dictionary to find matches.  The
way that it works is heavily dependent on the way that the search order
concept is implemented, the way that words are linked together in the
dictionary, and the way that hashing, if any, is done.
Again, it shouldn't be hard to implement on most systems, but it
can't be done as a `standard' program.
.PP
The code which matches substrings in the dictionary with the search string
depends on the way that names are stored in the dictionary, with tag bits
and such.  This code assumes that names are stored with tag bits at the
beginning and end, and the initial length byte may contain an immediate bit.
Considering the trouble that dictionary maintenance code must go to in order
to cope with these tag bits, I would argue that the space saved is not worth
the trouble. I think Forth implementations should use a single consistent
format for strings, and that format should be used for names in the
dictionary, with no tag bits.
.PP
I don't think it is possible to implement command completion at all on
systems that use the `first three letters and a count' scheme for saving
dictionary space.  Also, some sorts of hashing may make it harder.  The
hashing function in F83 just uses the first character of the name to do
four-way hashing, which is very convenient for command completion purposes.
If the hashing function were to require that the entire name be available to
calculate the hash code, it would be necessary to ignore the hashing and
simply search the whole dictionary.
.SH
Terminal support
.PP
The editor uses some `smart-terminal' features for manipulating the screen
display.  There may still be some terminals around that don't support the
necessary fuctions, but nearly all modern ones do.  Code is provided for
the Heath/Zenith H-19 command sequences and for the ANSI standard command
sequences.  It should be trivial to add the code for your favorite terminal.
.SH
Future Enhancements
.PP
There are some Emacs features that ought to be added to this editor.
Notably, it should be possible to set a marker at some position in the line,
move the cursor, and then either delete or copy the section of the line
between the marker and the cursor.  This wouldn't be very hard to implement.
Slightly harder would be command repetition counts, allowing the user to
type a number representing how many times a command should be executed.
.PP
Most of the time in a screen-oriented text editor is spent editing within a
line.  This means that this editor could serve as the basis for a complete
full-screen text editor, basically Emacs-in-Forth.  A really complete Emacs
implementation would require a lot of work, though, because Emacs usually
has the ability to split the screen into a number of non-overlapping windows,
editing a different file in each window.
.SH
Acknowledgements
.PP
I owe a debt of gratitude to the people who designed and evolved Emacs,
for providing an excellent model for how an editor should behave.  It is my
feeling that the hard part of program design is deciding what the program
SHOULD do, and that the implementation is usually trivial once the
specification is right.  In this case, I was absolutely convinced that the
Emacs editing model was exactly right; as a consequence, the implementation
of the editor other than command completion was completed in one evening.
.PP
I'm also indebted to the many people who have evolved the command completion
concept, so that I was able to just do the implementation, without having to
design the user interface.
.PP
Finally, I'm indebted to Mike Perry and Henry Laxen for the fine job they
did with F83.  Previous versions of Forth that I have used have lacked
many necessary (to me at least) features, so that I was forced to make
extensive modifications to them.
I hesitated to publish code for fear of having to explain how my
kernel works.  F83 does so many things right that I have been able to
write a lot of
applications code without having to spend too much time fixing the kernel.
.SH
References
.LP
[STAL] Stallman, R. M. \fIEMACS Manual for ITS Users\fP.  AI Lab Memo 554,
MIT, Cambridge, Mass., (1980).
.LP
[BOBR] Bobrow, D. G. Tenex: a paged time-sharing system for the PDP-10.
\fICommunications of the ACM\fP 15, 3, March 1972, p135-143.
.LS
.LG
( Forth 83 )
( Terminal control. Not all desireable functions are implemented )

only forth also definitions

defer cursor-backward
defer cursor-forward
defer insert-char-on-screen
defer delete-char-on-screen
defer kill-to-eol-on-screen

vocabulary terminals
terminals definitions

27 constant esc

: esc-prefix      esc emit   ;
: h19-cursor-forward  esc-prefix   ascii C emit  ;
: h19-cursor-backward esc-prefix   ascii D emit  ;
: h19-insert-char-on-screen ( char -- )
  esc-prefix ascii @ emit
  emit
  esc-prefix ascii O emit
;
: h19-delete-char-on-screen  esc-prefix   ascii N emit  ;
: h19-kill-to-eol-on-screen  esc-prefix   ascii K emit  ;
: h19
  ['] h19-cursor-forward  (is cursor-forward
  ['] h19-cursor-backward (is cursor-backward
  ['] h19-insert-char-on-screen (is insert-char-on-screen
  ['] h19-delete-char-on-screen  (is delete-char-on-screen
  ['] h19-kill-to-eol-on-screen  (is kill-to-eol-on-screen
;
: esc-prefix
  esc emit ascii [ emit
;
: ansi-cursor-forward  esc-prefix   ascii C emit  ;
: ansi-cursor-backward esc-prefix   ascii D emit  ;
: ansi-insert-char-on-screen ( char -- )
  esc-prefix ascii @ emit
  emit
;
: ansi-delete-char-on-screen  esc-prefix  ascii P emit  ;
: ansi-kill-to-eol-on-screen  esc-prefix  ascii K emit  ;

: ansi
  ['] ansi-cursor-forward  (is cursor-forward
  ['] ansi-cursor-backward (is cursor-backward
  ['] ansi-insert-char-on-screen (is insert-char-on-screen
  ['] ansi-delete-char-on-screen (is delete-char-on-screen
  ['] ansi-kill-to-eol-on-screen (is kill-to-eol-on-screen
;
: esc-prefix
  esc emit
;
: t925-cursor-forward   control L emit  ;
: t925-cursor-backward  control H emit ;
: t925-insert-char-on-screen ( char -- )
  esc-prefix ascii Q emit
  emit
;
: t925-delete-char-on-screen  esc-prefix  ascii W emit  ;
: t925-kill-to-eol-on-screen  esc-prefix  ascii T emit  ;

: t925
  ['] t925-cursor-forward  ['] cursor-forward  (is
  ['] t925-cursor-backward ['] cursor-backward (is
  ['] t925-insert-char-on-screen ['] insert-char-on-screen (is
  ['] t925-delete-char-on-screen  ['] delete-char-on-screen  (is
  ['] t925-kill-to-eol-on-screen  ['] kill-to-eol-on-screen  (is
;

: televideo t925 ;
: heath h19 ;
: sun ansi ;
  
( Emacs-style line editing for command lines.  Needs to be more )
( efficient at updating the screen, rather than doing everything )
( a character at a time.  See `install-line-editing' for functions )
( implemented and key bindings )

( To run this successfully under Unix, you have to )
( "stty nl cbreak -echo" )

( requires terminals.f )
ansi
only forth also definitions

decimal

  ( maximum size of input buffer )
variable buflen
  ( address of start of input buffer )
variable bufstart
  ( number of characters currently in input buffer )
variable bufcurrent
  ( character that was most recently typed, or character to add )
variable lastchar
  ( position of cursor within line )
variable bufcursor   0 bufcursor !
  ( remembers size of previous line )
variable last-bufcurrent  0 last-bufcurrent !

only forth also definitions
: forward-in-buf
  bufcursor @  bufcurrent @ <
  if   1  bufcursor +!
  else beep
  then
;
: forward-character
  forward-in-buf
  cursor-forward
;
: backward-character
  bufcursor @
  if   -1 bufcursor +!
       cursor-backward
  else beep
  then
;
: erase-next-character
  bufcurrent @
  if  -1  bufcurrent +!
       bufstart @  bufcursor @   +   dup 1+  swap
       bufcurrent @  bufcursor @  -  ( # of chars to move )
       cmove
       delete-char-on-screen
  else beep
  then
;
: erase-previous-character
  backward-character
  erase-next-character
;
: insert-character
  bufcurrent @  1+   buflen @  min   bufcurrent !
  bufstart @  bufcursor @   +   dup dup 1+  ( addr addr addr+1 )
  bufcurrent @  bufcursor @ - 1-  cmove>    ( addr )
  lastchar c@  swap  c!  (  )
  lastchar c@ insert-char-on-screen   (   )
  forward-in-buf
;
: beginning-of-line
  bufcursor @ 0
  ?do   backward-character  loop
;
: end-of-line
  bufcurrent @  bufcursor @  -  0
  ?do   forward-character   loop
;
: finish-line
  end-of-line
  bufcurrent @ last-bufcurrent !
  bufcurrent @ span !
  buflen @ bufcurrent !
  bl emit
  0 bufcursor !
;
: kill-to-end-of-line
  bufcursor @ bufcurrent !
  kill-to-eol-on-screen
;
: grab-last-line
  last-bufcurrent @ dup  bufcurrent ! bufcursor !
  bufstart @  bufcurrent @  type
;
: retype-line
  cr
  bufstart @  bufcurrent @  type
  bufcurrent @ bufcursor @  - 0
  ?do  cursor-backward loop
;
: erase-blanks-backward ( -- )
  bufcursor @
  if
    bufstart @ bufcursor @ 1- bounds swap   ( bufstart bufend )
    do  i c@
         bl <> if leave then
         erase-previous-character
    -1 +loop
  then
;
: erase-non-blanks-backward ( -- )
  bufcursor @
  if
     bufstart @ bufcursor @ 1- bounds swap   ( bufstart bufend )
     do  i c@
         bl = if leave then
         erase-previous-character
     -1 +loop
   then
;
: erase-previous-word 
  erase-blanks-backward
  erase-non-blanks-backward
;
: erase-blanks-forward ( -- )
  bufcurrent @ bufcursor  @
  ?do  bufstart @ bufcursor @ + c@
       bl <> if leave then
       erase-next-character
  loop
;
: erase-non-blanks-forward ( -- )
  bufcurrent @     bufcursor  @
  ?do  bufstart @ bufcursor @ +  c@
      bl = if leave then
      erase-next-character
  loop
;
: erase-next-word 
  erase-blanks-forward
  erase-non-blanks-forward
;
: beginning-of-word
  bufcursor @
  if
    bufstart @ bufcursor @ 1- bounds swap   ( bufstart bufend )
    do  i c@
         bl = if leave then
         backward-character
    -1 +loop
  then
;
: end-of-word ( -- )
  bufcurrent @ bufstart @ +
  bufcursor  @ bufstart @ +
  ?do i c@
      bl = if leave then
      forward-character
  loop
;
: forward-word
  forward-character
  end-of-word
;
: backward-word
  backward-character
  beginning-of-word
;
: quote-next-character
  key lastchar c!  insert-character
;

only forth also definitions
vocabulary keys-forth

variable keys  ' keys-forth   keys !

create name-buf 32 allot

: add-char-to-string ( str char -- )
  over ( str char str )
  count dup >r ( str char addr len )
  + c!  ( str )
  r> 1+ swap c!
;
: add-char-to-name ( str char -- )
  dup bl <
  if   ( control character so translate to ^ form )  ( str char )
     over ascii ^ add-char-to-string  ( str char )
     ascii A 1- +  ( str char' )
  then
  add-char-to-string
;

( searches the context vocabulary and nothing else )
: vfind ( str voc-threads -- str false | cfa true )
  >body >user over swap  hash @ (find
;

: do-command ( -- )
  name-buf lastchar c@  add-char-to-name
  name-buf keys @ vfind ( cfa true | addr false )
  if  execute else beep then
;

: (newexpect   (s adr len -- )
   dup span ! buflen !  bufstart !
   0 bufcurrent !  0 bufcursor !
   begin   bufcurrent @  buflen @ < 
   while   key  dup  lastchar c!  dup  bl <
     if    drop   name-buf width @ blank   0 name-buf !   do-command
     else  127 =
        if   erase-previous-character   else   insert-character    then
     then
   repeat
;

only forth also keys-forth definitions

: ^F forward-character ;
: ^B backward-character ;
: ^A beginning-of-line ;
: ^E end-of-line ;
: ^D erase-next-character ;
: ^H erase-previous-character ;
: ^K kill-to-end-of-line ;
: ^M finish-line ;
: ^J finish-line ;
: ^Q quote-next-character ;
: ^L retype-line ;
: ^[ key lastchar c! do-command ;
: ^[h erase-previous-word ;
: ^[d erase-next-word ;
: ^[f forward-word ;
: ^[b backward-word ;
: ^[= grab-last-line ;

( Command completion package a la TENEX. )

  ( index of char at the beginning of the latest word in the input buffer )
variable start-of-word

variable #candidates   0 #candidates !

: word-to-string ( -- str )
  bufstart   @  start-of-word @  +  ( addr of start of word )
  bufcursor  @  start-of-word @  -  ( start-addr len )
  'word  place
  'word
;
: collect-string ( -- str )
  ( finds start of this word and the current length of the word )
  ( leaves address of a packed string which contains that word )
  end-of-word
  bufcursor @    start-of-word !
  bufcursor @
  if
       bufstart @  bufcursor @  1-  bounds  ( bufend bufstart )
       swap  ( bufstart bufend )  ( loop runs backwards over buffer )
       do i c@ bl = if leave then
          -1 start-of-word +!
       -1 +loop
  then
  word-to-string  ( str )
;
: add-space ( -- )  bl lastchar c! insert-character ;

( leaves address of temporary buffer space for collecting match candidates )
: candidates  ( -- addr ) pad 100 + ;

: end-of-name? ( nfa+n -- nfa+n f )
  dup c@  128 and
;
: substring?  ( addr nfa -- f )

  2dup c@ width @ and swap c@   ( addr nfa nlen tlen )
  < if 2drop false exit then  ( not a substring if string longer than name )

  true -rot  ( true addr nfa )
  1+ swap    ( true nfa+1 addr )
  count bounds
  ?do   ( f nfa+n )
      dup c@  127 and  ( f nfa+n char-in-name )
      i c@ <>   ( f nfa+n character-mismatch? )
      if   nip  false swap  leave   then
      1+
  loop  ( f nfa+m )
  drop
;
: new-candidate ( lfa -- lfa )
  dup l>name  ( lfa nfa )
    candidates #candidates @ 2* +   !   ( lfa )
  1 #candidates +!   ( lfa )
;
: find-candidates-in-thread ( addr thread -- addr )
  begin  dup ( addr lfa lfa )
  while  2dup l>name  substring?  ( addr lfa f )
         if new-candidate then
       @
  repeat
  drop  ( addr )
;
: find-candidates-in-voc ( addr voc -- addr )
  over swap hash @ ( addr thread )
  find-candidates-in-thread
;

: find-candidates ( str -- )
  #candidates off
  prior off        ( addr )
  #vocs 0
  do
     context i 2* +   @   dup         ( addr voc voc )
     if   dup prior @ over prior !  = ( addr voc prior=voc? )
          if   drop                   ( addr )
          else find-candidates-in-voc ( addr )
          then
     else drop                        ( addr )
     then
  loop
  drop
;
: cclash? ( char# char str -- char# char f )
  ( if "char" is different from the "char#"-th character in str, leave true )
  count width @ and    ( char# char str+1 count )
  3 pick <=  ( char# char str+1 f )
  if   drop true     ( str too short is a clash )
  else 2 pick +      ( char# char address-in-string )
       c@  127 and ( char# char str-char )
       over <>
  then
;
: candidates-agree? ( char# -- char true | false )
  candidates @       ( char# first-candidate-nfa )
  2dup c@ width @ and >=         ( char# first-nfa f )

( if the test string is the same length as the first candidate, )
( then the first candidate has no char at position char#, so there )
( can be no agreement.  Since the test string is a substring of all )
( candidates, the > condition should not happen )

  if   2drop false  exit then  ( char# false )
  true -rot   ( true char# first-candidate )
       over + 1+ c@ 127 and     ( f char# char )

  ( now test all other candidates to see if their "char#"-th character )
  ( is the same as that of the first candidate )

       candidates 2+  #candidates @  1-  2*  bounds
       ?do   ( f char# char )
            i @ cclash?  ( f char# char f )
            if  rot drop  false -rot  leave then
       2 +loop        ( f char# char )
       rot if   nip true   else   2drop false   then
;
: expand-initial-substring ( -- )
  bufcursor @  start-of-word @  -
  begin                          ( current-length )
         dup candidates-agree?   ( current-len [ char true ] | false )
  while  lastchar c! insert-character
         1+                      ( current-length )
  repeat
  drop
;
: one-candidate? ( -- f )
  ( true if there is only one candidate or if they all are the same name )
  true   ( prime the loop )
  candidates  #candidates @ 2*  bounds
  ?do  i @  count drop  ( f addr1 )
       candidates @ count width @ and  ( f addr1 addr2 len )
       compare   ( f mismatch? )
       if  drop false leave then
  2 +loop
;
: do-erase ( -- ) ( Side effect: bufcurrent and bufcorsor may be reduced )
  begin
    word-to-string   ( addr )
    find-candidates
    #candidates @ 0=
  while
    erase-previous-character 
  repeat
;
defer word-is-complete
' add-space is word-is-complete
: do-expand ( -- )
  expand-initial-substring
 ( If expansion does not result in a unique choice, beep )
  one-candidate?
  if  word-is-complete  else beep  then
;
: expand-word ( max-len bufstart current char -- ... )
              ( ... --- max-len bufstart new-current' char )
  collect-string   ( str )
  find-candidates  ( )
  #candidates @
  if   do-expand   else  do-erase  then
;
: show-candidates  ( -- )
       candidates #candidates  @  2* bounds
       ?do  i @ .id  2 +loop
;
: do-show ( max-len bufstart current-cnt char -- ... )
          ( ...  max-len bufstart current-cnt char )
  cr
  collect-string   ( str )
  dup c@
  if   find-candidates 
       show-candidates
  else drop ." Any word at all is a candidate." cr
            ." Use words to see the entire dictionary"
  then
  retype-line
;

only forth also keys-forth definitions

: ^J expand-word ;
: ^_ do-show ;

only forth also definitions
.SM
.LE
