.nr PS 11
.nr ps 11
.TL
Operating System File Interface
.AU
Mitch Bradley
.AB
Report of the Operating Systems Working Group from the
1984 Rochester Forth Applications Conference.
.AE
.SH
Introduction
.PP
The Operating Systems working group at the 1984 Rochester Forth Applications
Conference discussed the desireability of defining a set of Forth words
to be used to access common operating system services.  A goal was to define
words that do not depend on which particular operating system is involved.
The general feeling was that there are some types of services which are
unique to particular operating systems, but that there is a large class
of services that nearly every system provides in basically the same way.
.SH
Services Considered
.PP
Several classes of operating system services were discussed:
.IP "File I/O"
Read, Write, Open, Close, Create.
.IP "Display I/O"
Smart Terminal handling, X/Y addressing, Graphics.
.IP "Multitasking, Multiuser"
Creating processes, Inter-process communication, Synchronization.
.IP "Other"
Passwords, Protection, Overlays, Dynamic Loading, Time, Networking.
.PP
The decision was made to focus the rest of the meeting on the file
system interface.
.SH
File System Interface
.PP
The following scheme for accessing files was agreed-upon.
The scheme is quite similar to a proposal published by Bradley
in the 1983 FORML Conference proceedings[1].
The working group made a number of improvements to the error handling
specified by the published paper.  The following is an excerpt from that
paper, edited to reflect the working group's recommendations.
.SH
Excerpt from Reference [1]
.PP
A file is viewed as a sequence of bytes.
The bytes may be accessed either sequentially or randomly.
Any file may be accessed either way, and
sequential and random accesses may be arbitrarily interspersed.
There is no concept of a "record"; applications which need such an
abstraction may easily provide it themselves using the primitives which
exist.  This philosophy is modeled after the 
.UX
[2] file system.
.SH
File Names
.PP
Since the file system is intended to be compatible with many different
operating systems, and since no two operating systems use the same naming
conventions for files, the file system imposes no restrictions of its own
on file names.  A file name is simply a counted string in the usual Forth
sense.  The string may contain any sequence of characters that is a legal
name for whatever native operating system is being used.  This presents
a minor portability problem when moving an application between operating
systems.  My experience with porting other code between dissimilar operating
systems shows that this is not a terrible problem.
.PP
The principle of carefully specifying the the set of words that are used
to access the bytes of a file, while leaving unspecified the details of
the file system naming and hierarchy, has been used successfully before.
The C language "standard I/O" [3] library and the primitive routines used
in the "Software Tools" [4] package both use this this concept.  Both of these
systems have been ported to a large number of different machines and operating
systems.
.SH
Addresing
.PP
The bytes within a file may be thought of as being numbered from 0 up to
one less than the number of bytes in the file.  For random accesses to
files, the address is specified with each access.  The address is a
32-bit number, allowing files to be up to 2**32 bytes (about 4* 10**9).
If the native operating system you are using doesn't support files this big,
then the operating system's maximum file size becomes the limiting factor.
A file may be thought of as an alternate address space,
allowing things like meta-compilation to a file.
.SH
High Level Interface Glossary
.PP
A "file descriptor" is a number that is used to refer to a file once
the file has been opened.
A file descriptor is denoted by "fd" in stack diagrams.
.PP
A "file name" is a packed string which is the ascci representation of
the name of a file.  See the section entitled "File Names".  File names
are passed around on the stack as the address of the packed string.
The stack diagram representation is "name".
.PP
The "file pointer" is the number of the byte which will be returned by the
next sequential access to the file.
.IP "OPEN ( name mode --- fd )"
.br
Given the name of a file, return a file descriptor which may be used to
subsequently access the file.  "Mode" specifies what sorts of access may be
performed on the file.  Legal values for mode are READ , WRITE ,
and MODIFY (these names are defined as CONSTANTs).  If the
file is not accessible, for example it doesn't exist, then 0 will
be returned as the fd value.
A successful open will set the file pointer to 0, so that the next
sequential access will occur at the beginning of the file.
.IP "CLOSE ( fd --- )"
.br
Stop using the given file descriptor to access the file that it is
currently associated with.  Since there is a limited number of file
descriptors, a file should be CLOSEd when it is no longer being used, so
that the file descriptor may be reused.
.IP "FPUTC ( byte fd --- )"
.br
Put the byte on the file associated with "fd" at the current file pointer
position.  Then increment the file pointer.
.IP "FGETC ( fd --- byte )"
.br
Get the byte from the file associated with "fd" which is at the current
file pointer position.  Then increment the file pointer.  If there are
no more bytes in the file, return EOF (-1) instead.  Note that any byte
is in the range 0..255, whereas EOF is -1, so they are distinct.
.IP "FPUTS ( addr count fd --- )"
.br
Put "count" bytes on the file associated with "fd".  Bytes are taken from
memory starting at address "addr", and are put on the file starting
at the current file pointer position.  The file pointer is advanced to
point to the byte just after the last byte put on the file.
.IP "FGETS ( addr count fd --- nread )"
.br
Get at most "count" bytes from the file associated with "fd".  Bytes are
read from the file starting at the current file pointer position, and
are put in memory starting at address "addr".  FGETS returns the number
of bytes actually read, which may be less than count, but not more.
If FGETS returns 0, it means that the file pointer was already at the
end of the file.
.IP "FILEC! ( byte d.addr fd --- )"
.br
Store the byte into the file associated with "fd" at the position indicated
by "d.addr".  "d.addr" is a 32-bit number.  The file pointer is moved to
d.addr+1.
.IP "FILEC@ ( d.addr fd --- byte )"
.br
Fetch a byte from the file associated with "fd" from the position indicated
by "d.addr".  "d.addr" is a 32-bit number.  The file pointer is moved to 
d.addr+1.
.IP "FSEEK ( d.addr fd --- )"
.br
Move the file pointer for the file associated with "fd" to the position
indicated by "d.addr".  "d.addr" is a 32-bit number.
.IP "FTELL ( fd --- d.addr )"
.br
Return the current value of the file pointer for the file associated with
"fd".
.IP "GETWORD ( addr fd --- addr )"
.br
Get the next sequence of non-white charactersfrom the file associated with
"fd".  The characters will be stored as a counted string in memory starting
at address "addr".  A "non-white" character is any character except blank,
tab, newline, or carriage return.  Leading "white" characters are skipped.
.IP "GETCWORD ( addr delim fd --- addr )"
.br
Get the next sequence of non-delimiter characters from the file
associated with
"fd".  The characters will be stored as a counted string in memory starting
at address "addr".  Leading delimiters are NOT ignored.  If the next 
character in the file in a "delim", then the null string (count=0) will be
left at addr !
.IP
The behavior of GETWORD and GETCWORD with respect to leading
delimiters is consistent with the way that they are actually used in practice.
Specifically, BL WORD is used to scan for the next Forth word, and it is
desirable to ignore leading blanks.  Tabs, newlines, and carriage returns
do not appear in Forth screens, so it is not necessary to deal with them.
They do appear in files created by other operating systems, so a file system
which attempts to be compatible with other operating systems must also
deal with them. When WORD is used with some delimiter other than BL, the
intention is really that leading delimiters should NOT be ignored!  For
example, when WORD is used to scan for the ")" which terminates a comment,
it is desireable to find one immediately in case of the null comment ( ).
Similarly, ."  and ABORT" should work with null strings, as in ." "
ABORT" "   .  The usual work-around is to explicitly check the
input stream to see if a delimiter is there before calling WORD.  This just
points to the fact that there should be two different versions of WORD,
one which gets non-white characters, ignoring leading white-space, and
one which takes an explicit delimiter off the stack and terminates as soon
as it encounters a delimiter.
.IP
The actual delimiter which terminated the GETWORD or GETCWORD is stored in
the global variable DELIMITER.  If an end-of-file condition terminated the
operation, the value of DELIMITER is EOF.
.IP "MAKE ( name mode --- )"
.br
Create a new file whose name is "name".  If the file already exists,
the operation will fail.
.IP "DELETE ( name --- f )"
.br
Get rid of the file whose name is "name".
.IP "FILES ( --- )"
.br
Show the names of the files in the current directory (whatever that means!).
.SH
End of Excerpt 
.SH
Error Reporting
.PP
Errors are reported with the word FILE-ERROR.
.IP "FILE-ERROR ( -- error-code )
.br
FILE-ERROR returns an error-code if the most recent file operation failed,
or 0 if it succeeded.  Each file system operation resets the error-code
value returned by FILE-ERROR.
.SH
Implementation
.PP
It should be noted that FPUTC, FGETC, FILEC!, and FILEC@ may be easily
implemented in terms of the FGETS and FPUTS.  However, such an
implementation is likely to be very slow.  An efficient implementation
is likely to require some sort of buffering.  The paper on which this
proposal is based included the code for efficient implementations for
CP/M-80 and for 68000 Unix.
The machine dependent parts were carefully factored,
so it is believed that adapting the code to other systems will be easy.
In particular, all the buffering is system-independent.
.PP
The published code does not exactly conform to the above proposal, due
to the recent changes in error handling, but it is quite close.
The author intends to publish revised code in the near future.
.SH
Working Group Members
.IP "Mitch Bradley (Chairman)"
.IP "David Lindbergh (Secretary)"
.IP "Bill Sebok"
.IP "Gary Nemeth"
.IP "Larry Hamilton"
.IP "Arne Henden"
.IP "Paul Conaway"
.IP "Jim Gaston"
.IP "Dennis Ruffer"
.IP "Tony Sager"
.IP "John Montgomery"
.IP "Barbara Zimmerman"
.IP "Walter Pawley"
.IP "Thomas Rens"
.SH
References
.IP 1.
Bradley, W. M., "A Portable File System Interface for Forth"
and "Implementations of a Portable File System Interface",
\fIProceedings of the 1983 FORML Conference\fP.
.IP 2.
Ritchie, Dennis, and Thompson, Ken. "The UNIX Time Sharing System",
\fICommunications of the ACM\fP, Vol. 17 No. 7 (July 1974).
Also see the October 1983 issue of \fIByte\fP.
.IP 3.
Kernighan, Brian, and Ritchie, Dennis. \fIThe C Programming Language\fP.
Prentice-Hall, 1978.
.IP 4.
Kernighan, Brian, and Plaugher, Bill. \fISoftware Tools\fP.
Prentice-Hall.
