.nr HM .3
.ds LF W. M. Bradley
.ds RF Page % of 16 pages
.ds CH
.TL
A Portable File System Interface for Forth
.AU
W. M. Bradley
.AI
302 Easy St. #29
Mountain View, CA 94043
(415) 961-1302
.AB
A file system interface for Forth has been designed and implemented.
This interface may be used for Forth systems which run under almost
any native operating system, or for stand-alone Forth systems.
Native operating system files are used where available.
The words which the program uses to access files are the same regardless
of the native operating system.
The interface is simple yet general.
Performance is excellent; loading from files is roughly the same speed as
loading from screens, and sometimes faster.
This work is placed in the public domain.
.AE
.bp
.LP
.bp
.SH
Abstract
.PP
A file system interface for Forth has been designed and implemented.
This interface may be used for Forth systems which run under almost
any native operating system, or for stand-alone Forth systems.
Native operating system files are used where available.
The words which the program uses to access files are the same regardless
of the native operating system.
The interface is simple yet general.
Performance is excellent; loading from files is roughly the same speed as
loading from screens, and sometimes faster.
This work is placed in the public domain.
.SH
Motivation
.PP
The need for a file system for Forth is often discussed.
Some of the advantages of files over BLOCKs are:
.IP o
Remembering which BLOCKs make up an application is painful even when the
storage medium is a 200 Kbyte floppy.  If you are using a 10 Mbyte hard disk,
it is nearly impossible.  File systems let you remember a name instead of
several numbers.
.IP o
Making backups with BLOCKs is difficult, because the backup medium often
does not have the same set of block numbers free as the original medium.
File systems do not have this problem.
.IP o
Portability of source code is reduced, because the load screens reference
block numbers, which usually have to be changed when moving the application
to another machine.
.IP o
While 1K blocks tend to constrain definitions to a reasonable length,
they also impose an artificial boundary that can lead to hard-to-read
code.  This can occur when you find yourself having to cram a definition
onto an almost-full screen.  The first thing to go is often the spacing that
contributes to readable structure, followed by the stack comments.
.IP o
BLOCKS waste disk space because screens usually contain a large percentage of
trailing blanks which pad the line lengths out to C/L characters.
.IP o
Most other operating systems use files.  It is difficult to share data between
Forth blocks and the files produced by other operating systems.
.SH
Reinventing the Wheel
.PP
A number of proposals for file systems have appeared in the literature
[1,2,3,4,5,9].  Why should I invent yet another file system?  The file
systems that I have seen don't meet all the requirements that I consider
mimportant.
.IP o
While it is theoretically nice to define a wonderful file system for Forth
that runs without any other "inferior" operating system, in practice other
operating systems exist and co-exist with Forth.  It is very useful to be
able to share files with an operating system that already exists on your
machine.  For this reason, the Forth file system should use the same files
as the "other" operating system, whatever that happens to be.  This
disqualifies file systems which "start from scratch" (I don't mean to
belittle the contributions of others in defining nice file systems, I
just think that the ability to use other systems' files is too valuable to
give up).
.IP o
Despite the use of the other operating systems' files, the words that Forth
uses to access those files should not vary from system to system.
File system interfaces which just allow you to execute the native operating
system's system call from Forth lose on this count.
.IP o
Some "file systems" are just interfaces to operating system files so that
BLOCKs may actually be stored in files.  This is useful but doesn't solve
most of the problems presented in the previous section.
.IP o
The file system should be available to everybody.  In the context of Forth,
this means PUBLIC DOMAIN code!
.SH
So here it is ...
.PP
I present for your consideration a file system interface which meets
my requirements.  It is in the public domain.  It can run on top of virtually
any operating system and use that operating system's native files.
The words that a Forth application uses to access those files are the
same no matter which operating system is the base.  A compatible stand-alone
version is possible, although I haven't implemented that.  The performance
is very good - loading from a file is approximately the same speed as
loading from screens.  In some cases the file loads faster than the screens!
.PP
An additional benefit is device independence.  The exact same words that
are used to access conventional files may also be used to access other I/O
devices, such as printers, modems, terminals, the console, etc.
.PP
This system has been ported to a Z-80 CP/M\(dd system and a 68000
.UX
system.
I think that it is easily implementable for just about any existing
operating system.
.FS
\(dd CP/M is a Trademark of Digital Research.
.FE
.SH
Philosophy
.PP
A file is viewed as a sequence of bytes.
The bytes may be accessed either sequentially or randomly.
Any file may be accessed either way, and
sequential and random accesses may be arbitrarily interspersed.
There is no concept of a "record"; applications which need such an
abstraction may easily provide it themselves using the primitives which
exist.  This philosophy is modeled after the 
.UX
[6] file system.
The simple yet sufficient file model is one of the reasons why
.UX
 has achieved such popularity.
.PP
Some operating systems (e.g. CP/M) insist on imposing a record structure
on their files.  My file system copes with this, and converts it to
a byte-oriented interface so that the Forth programmer need not worry about
the record structure imposed by the underlying operating system.
.PP
The interface to this file system is specified at two different levels.
The most important interface (called the "high-level interface"),
the only one that the applications programmer sees, is the set of
WORDs that a program uses to access files.  The other interface (the
"low level interface") is the set of WORDs that the file system uses to
actually access the data from whatever native operating system is being
used.  It is this low-level interface which makes it possible to easily
port the file system to different native operating systems.
.SH
File Names
.PP
Since the file system is intended to be compatible with many different
operating systems, and since no two operating systems use the same naming
conventions for files, the file system imposes no restrictions of its own
on file names.  A file name is simply a counted string in the usual Forth
sense.  The string may contain any sequence of characters that is a legal
name for whatever native operating system is being used.  This presents
a minor portability problem when moving an application between operating
systems.  My experience with porting other code between dissimilar operating
systems shows that this is not a terrible problem.
.PP
The principle of carefully specifying the the set of words that are used
to access the bytes of a file, while leaving unspecified the details of
the file system naming and hierarchy, has been used successfully before.
The C language "standard I/O" library [7] and the primitive routines used
in the "Software Tools" [8] package both use this this concept.  Both of these
systems have been ported to a large number of different machines and operating
systems.
.SH
Addresing
.PP
The bytes wihtin a file may be thought of as being numbered from 0 up to
one less than the number of bytes in the file.  For random accesses to
files, the address is specified with each access.  The address is a
32-bit number, allowing files to be up to 2**32 bytes (about 4* 10**9).
If the native operating system you are using doesn't support files this big,
then it becomes the limiting factor.  A file may be thought of as an
alternate address space, allowing things like meta-compilation to a file,
etc.
.SH
Buffers
.PP
The file system code's primary job is to manage a set of buffers (which are
not necessarily the same as the set of buffers managed by BLOCK, etc.).
The buffers have two purposes.  First, they are necessary for converting
record-oriented operating system or disk interfaces to the byte-oriented 
interface provided by the file system.  Second, they are the means whereby
the performance of the system is kept high, since most accesses to a file
are satisfied by bytes in the buffer, thus reducing the number of times
that the operating system must be called to perform I/O operations.
The program using the file system need not know about the existence of
the buffering.  Particularly importantly, a Forth WORD may cross a buffer
boundary, so there is no need to make sure that source code lines up with
any boundaries.
.SH
File Descriptors
.PP
The primary "handle" that is used to access a file is the "file descriptor".
A file descriptor is actually the address of a memory area that contains
information about the file and about the buffer associated with it.  The
system manages a fixed number of file descriptors which are automatically
allocated and freed.
.SH
High Level Interface Glossary
.PP
A "file descriptor" is a 16-bit number that is used to refer to a file once
the file has been opened.
A file descriptor is denoted by "fd" in stack diagrams.
.PP
A "file name" is a counted string which is the ascci representation of
the name of a file.  See the section entitled "File Names".  File names
are passed around on the stack as the address of the counted string.
The stack diagram representation is "name".
.PP
The "file pointer" is the number of the byte which will be returned by the
next sequential access to the file.
.IP "OPEN ( name mode --- fd )"
.br
Given the name of a file, return a file descriptor which may be used to
subsequently access the file.  "Mode" specifies what sorts of access may be
performed on the file.  Legal values for mode are READ-ONLY , WRITE-ONLY ,
and READ-WRITE (these names are defined as CONSTANTs in the code).  If the
file is not accessible, for example it doesn't exist, then 0 is returned
instead of fd.  The file pointer is set to 0, so that the next sequential
access will occur at the beginning of the file.
.IP "CLOSE ( fd --- )"
.br
Stop using the given file descriptor to access the file that it is
currently associated with.  Since there is a limited number of file
descriptors, a file should be CLOSEd when it is no longer being used, so
that the file descriptor may be reused.
.IP "FPUTC ( byte fd --- )"
.br
Put the byte on the file associated with "fd" at the current file pointer
position.  Then increment the file pointer.
.IP "FGETC ( fd --- byte )"
.br
Get the byte from the file associated with "fd" which is at the current
file pointer position.  Then increment the file pointer.  If there are
no more bytes in the file, return EOF (-1) instead.  Note that any byte
is in the range 0..255, whereas EOF is a 16-bit -1, so they are distinct.
.IP "FPUTS ( addr count fd --- )"
.br
Put "count" bytes on the file associated with "fd".  Bytes are taken from
from memory starting at address "addr", and are put on the file starting
at the current file pointer position.  The file pointer is advanced to
point to the byte just after the last byte put on the file.
.IP "FGETS ( addr count fd --- nread )"
.br
Get at most "count" bytes from the file associated with "fd".  Bytes are
gotten from the file starting at the current file pointer position, and
are put in memory starting at address "addr".  FGETS returns the number
of bytes actually gotten, which may be less than count, but not more.
If FGETS returns 0, it means that the file pointer was already at the
end of the file.  This routine is not implemented yet.
.IP "FILE! ( byte d.addr fd --- )"
.br
Store the byte into the file associated with "fd" at the position indicated
by "d.addr".  "d.addr" is a 32-bit number.  The file pointer is moved to
d.addr+1.
.IP "FILE@ ( d.addr fd --- byte )"
.br
Fetch a byte from the file associated with "fd" from the position indicated
by "d.addr".  "d.addr" is a 32-bit number.  The file pointer is moved to 
d.addr+1.
.IP "FSEEK ( d.addr fd --- )"
.br
Move the file pointer for the file associated with "fd" to the position
indicated by "d.addr".  "d.addr" is a 32-bit number.
.IP "GETWORD ( addr fd --- addr )"
.br
Get the next sequence of non-white charactersfrom the file associated with
"fd".  The characters will be stored as a counted string in memory starting
at address "addr".  A "non-white" character is any character except blank,
tab, newline, or carriage return.  Leading "white" characters are skipped.
.IP "GETCWORD ( addr delim fd --- addr )"
.br
Get the next sequence of non-delimiter characters from the file
associated with
"fd".  The characters will be stored as a counted string in memory starting
at address "addr".  Leading delimiters are NOT ignored.  If the next 
character in the file in a "delim", then the null string (count=0) will be
left at addr !
.FS
N.B.  The behavior of GETWORD and GETCWORD with respect to leading
delimiters is consistent with the way that they are actually used in practice.
Specifically, BL WORD is used to scan for the next Forth word, and it is
desirable to ignore leading blanks.  Tabs, newlines, and carriage returns
do not appear in Forth screens, so it is not necessary to deal with them.
They do appear in files created by other operating systems, so a file system
which attempts to be compatible with other operating systems must also
deal with them. When WORD is used with some delimiter other than BL, the
intention is really that leading delimiters should NOT be ignored!  For
example, when WORD is used to scan for the ")" which terminates a comment,
it is desireable to find one immediately in case of the null comment ( ).
Similarly, ."  and ABORT" should work with null strings, as in ." "
ABORT" "   .  The usual work-around is to explicitly check the
input stream to see if a delimiter is there before calling WORD.  This just
points to the fact that there should be two different versions of WORD,
one which gets non-white characters, ingoring leading white-space, and
one which takes an explicit delimiter off the stack and terminates as soon
as it encounters a delimiter.
.FE
.IP "MAKE ( name --- f )"
.br
Create a new file whose name is "name".  If the file was created successfully,
return true.  Otherwise return false.  If the file already exists, return
false.
.IP "DELETE ( name --- f )"
.br
Get rid of the file whose name is "name".
.IP "FILES ( --- )"
.br
Show the names of the files in the current directory (whatever that means!).
.SH
Low Level Interface
.PP
These routines must be recoded for every different native operating system.
The higher-level routines call these routines through execution vectors
which are a part of the file descriptor structure.  This vectoring allows
the file system to associate different low-level routines with different
types of files.  This is the mechanism whereby device-independent I/O is
implemented.  For instance, CP/M provides totally different system calls
for accessing disk files and other devices like the terminal or the printer.
By vectoring these following routines, the file system can access any of
these devices transparently to the applications program.
.PP
.IP "ALIGN ( d.byteno --- d.aligned )"
.br
Round the byte number "d.byteno" down to the next lowest boundary which
must be enforced.  For instance, CP/M disk files are only accessible in
128-byte chunks, so any request to read a CP/M sector must start on a
multiple-of-128 boundary.  If the underlying system imposes no such
restrictions, this routine may just return d.aligned.  This is particualrly
useful for strictly-sequential devices like terminals and printers.
If the underlying system allows access to files at arbitrary byte positions
(this is the case for
.UX
, for example), it may still be desireable to
align requests to boundaries which are particularly efficient (e.g. - 512-byte
boundaries are good for
.UX
).  ALIGN is not called by the high-level code,
but the alignment concept is essential for implementing this file system
on record-oriented or block-oriented operating systems, so it is discussed
here.
.IP "FREAD   ( bflimit bfbase d.byteno fid --- bflimit' bfbase' d.aligned bftop bfend )
.br
Read a bufferfull of data from the file.  "fid" is the internal file
identifier.  "d.byteno" is the
byte that the upper-level code is really trying to get.  "bfbase" is
the lowest address in the buffer, and "bflimit" is one more than the
highest address in the buffer.  "fid" is the identifier used by the
underlying operating system to identify the file.  The "open" routine
(see below) discovers this number and stores it in the file descriptor
so that it may be passed to the low-level routines when they are called.
.IP
Returns "bfend", the highest address (plus one) in the buffer that is
available for writing.  This is nearly always equal to the address of the
end of the buffer, but may be set to the address of the start of the
buffer if the buffer becomes invalid for some reason (mainly when the
file is first opened).
"bftop" is one more than the highest address that contains actual
data read from the file.  If the read operation was unable to get enough
data to fill, the buffer, bftop may be less than bflimit.
"d.aligned" is the file address (the 32-bit byte number) which corresponds
to the byte which is at the buffer address "bfbase'".  "bfbase'" and
"bflimit'" are copies of the arguments.  They are returned in case the
FREAD routine wants to be sophisticated and return a different buffer
than the one passed to it.  This would be useful if you had a system which
kept a cache of disk blocks, so that accesses to a block which was already
in the cache didn't really cause any I/O, but instead just substituted the
appropriate buffer.
.IP "FWRITE ( bfend bftop bfbase fstart fid --- bfwrittend )"
.br
Write the bytes between bfbase and bftop-1 to the file specified by
fid.  Fstart is the file address which corresponds to the byte at
buffer address bfbase.  If the underlying system only allows data to
be written in fixed-sized chunks, then data up to address bfend may
be written.  bfend - bfbase will always be a multiple of the chunk-size
if FREAD works right.
Returns bfwrittenend, which should be >= bftop if everything goes well.
.IP "FCLOSE ( fid --- )"
.br
Do anything necessary to make sure everything is squared-away when the
file is no longer needed.  It is not necessary to worry about stuff that
may be left-over in the higher-level buffers; that will have already
been taken care of by the higher-level code.
.IP "FSIZE ( fid --- d.size )"
.br
Find out how many bytes are in the file.
.IP "OPEN ( name mode --- fd )"
.br
This is the hard one!  OPEN prepares the file whose name is "name" for
access in mode "mode", returning a file descriptor which is used for
subsequent access to the file.  If the attempt fails, 0 is returned.
"Mode" can be either READ-ONLY, WRITE-ONLY, or READ-WRITE.
.IP
OPEN is tricky, because it must do several things.  The first thing is
to find a file descriptor to use.  My implementation of this is rather
simple: there are a fixed number of file descriptors which are kept
in one place.  There is a routine FIND-FD which searches through the
array of file descriptors until if finds one which isn't in use.
(There is a field in the file descriptor which tells whether or not
in use, among other things).
.IP
Another thing that OPEN must do is to set up the execution vectors
for the proper access routines for the file.
The open routine is very important for providing device independence.
The way you implement device independence depends strongly on what you
actually have to do to access the particular devices.  In general,
you need a read routine and a write routine for each device that must
be accessed using its own special code.  The open routine may need to
look at the file name and decide which read and write routines to use.
CP/M is a good example here.  In CP/M, besides normal disk files there
are 16 special names that refer to I/O devices (e.g. CON:, LST:).  There
are several different system calls that must be used to access these
devices.  The OPEN routine for CP/M first checks to see if the file name
given to it matches one of these special names.  If so, the addresses
of appropriate READ and/or WRITE routines are stored in the execution
vectors in the file descriptor.  If the name doesn't match one of the
special names, then OPEN tries to open a disk file with the specified
name.
.IP
Finally, OPEN must setup the file in the proper mode.  This is done
by using special READ and WRITE routines.  If the mode is WRITE-ONLY,
the READ routine is set to NULLREAD, which basically does nothing.
If the mode is READ-ONLY, the WRITE routine is set to NULLREAD, which
causes an error condition.
.IP
Attempts to OPEN a non-existent file fail, returning 0.
.SH
File management routines
.PP
The following words are used for managing entire files.  No assumption
is made here about any sort of hierarchy of files which may be provided
by the underlying operating system.  In this respect, the precise
behavior of these words is system-dependent.  Many systems have a notion
of a "current directory".  If that is the case, these words should be
implemented so that the files affected are in the current directory
unless some other directory is explicitly specified.  Example:  CP/M
has a "current default drive".  Unless the drive name is specified as
part of the file name, the defualt drive should be assumed.
.IP "DELETE ( name --- f )"
.br
The file with the given "name" is removed from the disk.
Returns true if successful, false if unsuccessful.
.IP "CREATE ( name --- f )"
.br
A file is created whose name is "name".  Returns true if successful,
false if unsuccessful.
.IP "FILES ( --- )"
.br
Displays a list of files in the "current directory", whatever that means
for the underlying system. This word is not implemented in the source code
presented here.  It would be nice to specify a name as an argument,
so that "wildcarding" could be done.
(Actually, that probably should be a separate word).
.SH
Source Code
.PP
This code consists of a portable section and a system-specific section.
The system-specific code has currently been implemented for both CP/M
and 68000
.UX.
Only the portable section is shown in this paper.
A companion paper, "Implementations of a Portable File System Interface",
gives the system-specific code for both CP/M and 68000
.UX.
.PP
This code has no KNOWN functional errors.  It does have some rough edges
which I am currently removing, and it may have errors that I don't know
about yet.
.PP
The code is Forth-79.
A few of the words are defined in 8080 machine code.  In most cases,
the high-level definition is given too, commented out.  I apologize
for the cases where I do not give a high-level equivalent.
68000 versions of all the code words appear in the companion paper.
.SH
Efficiency and performance
.PP
An extremely important performance issue is the speed of loading from files.
As far I/O is concerned, the important thing is the efficiency of WORD.
The Forth word ENCLOSE does a very efficient job of speeding-up WORD
when loading from screens.
ENCLOSE makes an important simplifying assumption:  WORDs do
not cross BLOCK boundaries.  This assumption can not be made for files,
because from a users' perspective, there are no boundaries at all inside
the file.  The buffering is completely transparent.  The solution is to
write a code word which collects as much of a word as possible without
going outside the current buffer.  The high level word "GETWORD"
takes care of getting another bufferfull and calling the code word again
if the word being collected crosses a buffer boundary.  Most of the time
the code word only needs to be called once, since most words don't cross
a buffer boundary.
.PP
Files have one speed advantage over screens.  Since lines on screens
must be padded out to C/L characters, a significant portion of the
characters stored and processed are trailing blanks.  Lines in files
don't need trailing blanks, so the total number of characters that
must be handled is less.  Another improvement, though less significant,
is the ability to use TAB characters to replace sequences of blanks,
further reducing the number of characters that must be processed.
GETWORD treats blanks, tabs, carriage returns, and
line feeds as all being the same.  (It might even be a good idea to
treat ALL control characters as being equivalent to blanks for the
purposes of WORD).
.PP
The net result is that loading from files is roughly the same speed
as from screens, and may be somewhat faster or slower depending on
the percentage of blanks in the screens.
.SH
Compiling from files
.PP
Once you have a file system, it is a simple matter to rewrite the
text interpreter to use files instead of screens or the terminal.
It is especially easy if you have a vectored WORD.  All you have to
do is open the file and change WORD so that it executes NEWWORD
instead.  (NEWWORD just looks a the delimiter and calls GETWORD if
the delimiter is BL, or calls GETCWORD otherwise.)  It is easy to
write a word FLOAD which does this.  FLOAD can even be used inside
a file to "include" another file within the first (similar to screens
loading other screens).
.SH
Outstanding Issues
.PP
Improvements need to be made in a few areas.
.PP
Error reporting needs to be expanded.
.PP
The file system may need to take special action if the
user's console is opened for output, so that output does not get stuck
in the buffer when the user is expecting to see it now.  It is possible
to set the buffer size to some small number (like 1) if the output is
going to the console, but this is pretty slow.  Another alternative
is to flush the buffer whenever a carriage return is sent.  This problem
is looking for a clean solution.
.PP
An editor is needed!  Many people will want to use the file editor already
on their system.  Nevertheless, I think that a public-domain
file editor written in Forth is an important thing to have.  Any volunteers?
.PP
The low-level code for a stand-alone version needs to be written.
This could be built on top of BLOCK (in particular, see [4]).
Some decisions have to be made;
specifically, what does a file name look like? and what sort of hierarchy
do files fit into?  Many of the previous file system proposals attempt to
answers these questions.  The important point that I would like to make is
that these questions are not the most important ones as far as a useful
and portable file system is concerned.  The important thing is to standardize
on the set of words to access the bytes within a file, rather than how
files are named.
.PP
It is possible to implement BLOCK on top of this file system, using
a file to store an array of BLOCKs.  While this approach may have its
uses, it seems like a step backwards to me.
.PP
It is possible (and easy!) to write a set of low-level routines to make
the system memory look like a file.  This has interesting implications
for meta-compiling.
.SH
Disclaimer
.PP
There are no new ideas presented in this paper.  Rather, this file system
is an attempt to make available to the Forth community a file system that
is based on already-proven principles and which is implemented in an
efficient and portable fashion.
.SH
Acknowledgements
.PP
The design of this file interface owes much to
.UX,
and in particular to
the design of the "Standard I/O" library for the C language[7].  C Standard
I/O has proven to be efficient, portable, and easy to use, so I felt that
it would be a good model.  The system also benefited from discussions with
Christian Jacobi concerning a file system interface for Modula-2.
.SH
References
.IP 1.
James, John S.,  "Pyramid Files: A Proposed FORTH File System".  \fIProc. 1981
FORML Convention\fP.
.IP 2.
Rible, John and Dowling, Tom, "A File System in FORTH",  \fIProc. 1981 FORML
Convention\fP.
.IP 3.
Colburn, Don, "A Minimum Automated Contiguous File Allocation System for
FORTH-79 Environments",  \fIProc. 1981 FORML Convention\fP.
.IP 4.
Helmers, Peter, "File Naming System", \fIForth Dimensions\fP, Vol. II No. 3.
.IP 5.
Delwood, Donald, "A FORTH Based File Handling System", \fIForth Dimensions\fP,
Vol. IV No. 3.
.IP 6.
Ritchie, Dennis, and Thompson, Ken. "The UNIX Time Sharing System",
\fICommunications of the ACM\fP, Vol. 17 No. 7 (July 1974).
Also see the October 1983 issue of \fIByte\fP.
.IP 7.
Kernighan, Brian, and Ritchie, Dennis. \fIThe C Programming Language\fP.
Prentice-Hall, 1978.
.IP 8.
Kernighan, Brian, and Plaugher, Bill. \fISoftware Tools\fP.
Prentice-Hall.
.IP 9.
See the Proceedings of the 1981 Rochester FORTH Standards Conference
for several papers about file systems of various sorts.
