.nf
.bp
( File system interface for Unix(tm)

( N.B. Requires a word "syscall".   syscall takes an unspecified )
( number of stack arguments and calls the system routine whose )
( number is on the top of the stack.  Upon return, the arguments )
( are still on the stack, since syscall doesn't know how many there )
( and therefore can't get rid of them. )
( Since the Unix I am using uses 32-bit integers, )
( all arguments are double numbers )
(  )
( The value returned from the function may be accessed by the word )
( "retval", or "dretval" if the return value is 32-bits. )
(  )
( The word "cstr" may be used to convert a Forth style counted string )
( to a C style null-terminated string )

( System call numbers.  May be different on some versions of Unix! )

 5 constant u_open   6 constant u_close   3 constant u_read
 4 constant u_write 19 constant u_lseek  10 constant u_unlink

( modes for opening files )
0  constant u_read-only   1 constant u_write-only
2  constant u_read-write

8 base ! 
644 1000 or constant create_mode
( 644 is the protection code for rw-r--r-- protection )
( 1000 is the "create" bit )
decimal

( magic numbers for the lseek system call )
 0 constant u.l_set       2 constant u.l_fromeof

( efficient block size )
512 constant u.recsize  th fe00 constant u.recmsk

u.recsize 1 * constant u.bufsize

( buffer allocation )
#fds constant #fbufs
 
create fbufs  #fbufs u.bufsize * allot
( allocation strategy - make as many fbufs as there are fds. )
( then if the n-th fd is used, also use the n-th fbuf )
 
: (find-fbuf ( fd --- fbuf )
  fds - fdsize / u.bufsize * fbufs +
;

( translation between unix file modes and forth file modes  )
: >unixmode ( fmode --- unixmode )
  read-only  over = if drop u_read-only  else
  write-only over = if drop u_write_only else
  read-write over = if drop u_read-write else
                       drop u_read-only
                    then then then
;

: unixalign ( d.byteno --- d.aligned )
  ( rounds down to a block boundary. )
  ( This isn't necessary, but it speeds things up )
  swap u.recmsk and swap
;
.bp
: unix_close ( unix_fd --- )
  s->d u_close syscall 2drop
;

: unix_open ( unixmode name --- unix-fd )
  swap s->d rot cstr u_open syscall retval >r 2drop 2drop r>
;

: lseek ( whence d.byteno unix_fd --- d.byteno )
  >r >r >r s->d r> r> r> s->d  ( convert whence and unix_fd to longs )
  u_lseek syscall dretval
  >r >r 2drop 2drop 2drop r> r> ( get rid of arguments )
;

: unix_set_position ( d.byteno unix_fd --- )
    >r
    u.l_set 2over1 r>  ( d.byteno whence d.byteno unix_fd )
    lseek ( bytecount addr d.byteno d.byteno' )
    2over 2over d=
      if   2drop 2drop   ( bytecount addr )
      else ." lseek failed - got " d. ." tried " d. quit then
;

: unixfsize ( unix_fd --- d.size_in_bytes )
  >r u.l_fromeof 0. r> lseek ( d.eof_byte_no )
;
 
( unixfread )
: unixfread ( bflimit bfbase d.byteno unix_fd --- )
            ( --- bflimit' bfbase' d.aligned bftop bfend )
  >r
  unixalign                          ( bflimit bfbase d.aligned )
  2dup r@  unix_set_position         ( bflimit bfbase d.aligned )
  2over - 4 pick r@    ( bflimit bfbase d.aligned   bytecount addr unix_fd )
  read                 ( bflimit bfbase d.aligned   bytes_read )
  dup -1 = if   1 abort" unixfread failed " cr quit then
  4 pick  +                          ( bflimit bfbase d.aligned bftop )
  r> drop
  5 pick ( bfend is the same as bflimit for this case )
;

( unixfwrite )
: unixfwrite ( bfend bftop bfbase fstart unix_fd  --- bfwrittend )
  >r
  r@  unix_set_position         ( bfend bftop bfbase )
  swap over - over r@           ( bfend bfbase   bytecount addr unix_fd )
  write                         ( bfend bfbase   bytes_written )
  +                             ( bfend bfwrittenend )
  r> drop swap drop
;
 
( unixfclose setupfd setfcb )
: unixfclose ( fd --- )
  fid @ unix_close
;

: nullread ( bflimit bfbase daddr fid --- ... )
           ( ... --- newbflimit newbfbase newfstart newbftop newbfend )
  drop 3 pick  5 pick
;

: nullwrite ( bfend bftop bfbase fstart fid --- )
  1 abort" You tried to write to a read-only file!"
; 
.bp
: setupfd ( mode unix_fd buf fd )
  >r dup u.bufsize + swap   r@   usebuf
  ( unix_fd )  r@ usefid       r@ fid @ unixfsize   r@ usesize
  ( mode ) dup r@ fmode !
  ( mode ) dup read-only and
  if  ['] unixfread  else ['] nullread then ( mode readroutine )
  swap write-only and
  if  ['] unixfwrite else ['] nullwrite then ( readroutine writeroutine )
  ['] unixfclose ['] nodev
  r>  useroutines
;
 
( unixfcreate unixfdelete )
: unixfcreate ( name --- f ) ( 1 success, 0 failure )
  create_mode swap ( create_mode name )
  unix_open ( unix_fd )
  dup 0<
  if   drop  0
  else unix_close 1
  then
;

: unixfdelete ( name --- f ) ( 1 success, 0 failure )
  cstr u_unlink syscall retval 0= >r 2drop r>
;

: unixfopen ( name mode --- fd ) ( 0 if failure )
  dup >unixmode rot ( mode  unixmode name )
  unix_open ( unix_fd )
  dup 0<
  if   drop 0   ( open failed )
  else find-fd dup (find-fbuf swap dup >r ( mode unix_fd buf fd )
       setupfd r> ( fd )
  then
;
