.TH FORTH ?? "11 April 1985"
.SH NAME
forth \- Forth interpreter/compiler
.br
forth \- Forth dynamic loader
.br
files.exe \- Binary for Unix version of forth
.br
stand.exe \- Binary for Standalone version of forth
.SH SYNOPSIS
.B forth
[ binary-file ] [ \fB\-v\fP ] [ file ] [ \fB\-\fP ]
.br
	[ \fB\-s 'string of forth commands'\fP ] \fB.\|.\|.\fP
.SH DESCRIPTION
.LP
.I Forth
is the Sun Forth interpreter/compiler which executes programs
in the Forth language.  Forth is an interactive programming language
with a stack-oriented, reverse-polish syntax.  It is extensible;
programs are written by defining new subroutines ("words" in Forth
jargon) which have equal status with system-supplied words.
Words must be defined in terms of already-existing words.  As a word is
defined, it is incrementally compiled and added to the Forth system.
A Forth word is executed simply by typing its name.
.LP
The Sun Forth implementation allows new words to be compiled either
from the keyboard or from Unix files.  The command line optionally
specifies the name of the forth binary file (which must have a
".exe" suffix) to start with, and a list of other files to compile.
If no ".exe" file is specified, a default file is used.
The default file is named \fIfiles.exe\fP.  The environment variable
FPATH is consulted to find a list of directories in which to look for
\fIfiles.exe\fP. \fI/usr/local/lib/forth/files.exe\fP is the forth binary
file which contains the Forth-83 Standard words and a lot of utilities,
including a "Standard I/O"-like interface to the Unix file system.
Once the desired files have been compiled, the system can
(but does not have to) take further input from the keyboard.  At any
time, a binary image of the state of the forth system may be saved.
This new image (".exe" file) may be later used as the starting binary
file for further invocations of forth.
.LP
The program \fI/usr/local/forth\fP is really just a
dynamic loader for the Forth environment.  It serves two
purposes: 1) to provide the forth binary code with access to Unix system
calls and library routines  2) to concatenate the list of files which
serve as the input stream to the forth system.
.LP
Stand-alone programs are compiled under Unix.  The forth system then writes
out an "a.out" format file that can be booted to run stand-alone.
The stand-alone program runs in supervisor state and has access to
all the hardware resources of the Sun Workstation.  When compiling a
stand-alone program, the binary file \fIstand.exe\fP is used as a starting
point instead of \fIforth.exe\fP.  \fIStand.exe\fP includes precompiled
words to access the hardware and to do console I/O via the Sun ROM monitor.
The forth word which writes a stand-alone "a.out" format image is
\fIsave-stand\fP.
.LP
While \fIstand.exe\fP is compiling the stand-alone program, it
does its I/O using the normal Unix facilities.  The ROM monitor routines
are not used until the "a.out" format image is booted stand-alone.
When the system is running stand-alone, the interactive features are
still available.  The system may be further extended by typing in
new forth words from the keyboard.
.LP
There are several hundred Forth words in this system, comparable to the
number of library routines available to a C program.  Thus it is impractical
to list them all in this manual page.
The \fBSEE ALSO\f section lists references which describe the Forth language
and the various libraries of Forth words available in the Sun Forth
implementation.
.SH OPTIONS
.LP
The following options are interpreted by
.IR forth .
.TP 8
.B \-v
Verbose: causes forth to print the name of each file as it is being compiled
by the forth system.  This is useful in debugging if a large list of
files is being used.  It helps locate the file in which a compilation
error occurred.
.TP
.B \-s 'string of forth commands'
When this option is encountered, the next command line argument is
passed directly to the forth system as a command to execute.  Unless
the string consists of a single word, it must be quoted.
This is usually used in makefiles to tell the forth system to write
out an image of itself.  The argument to -s needs to be quoted unless
it consists of a single Forth word.  A typical use is
.br
-s '"" test.exe save-forth'
.br
which would case a file "test.exe" to be written.
.TP
.B \-
Whenever a minus sign by itself is encountered,
input is taken from the standard input.
This may occur more than once in the command list.
If no files are specified, the standard input is automatically used.
.SH ENVIRONMENT
.LP
The environment variable FPATH is used to determine this list of
directories to search for the forth binary image.  It is also used within
Forth to determine which directories to search during a PATH-OPEN ,
PATH-LOAD , or REQUIRES command.
.LP
Here is how you might want to set FPATH from inside a .login file:
.LP
setenv FPATH .:$HOME/forth:/usr/local/lib/forth
.LP
This specifies that the current directory is to be searched first, followed
by the users private forth directory, followed by the system forth library.
.SH EXAMPLES
.LP
To start the standard forth system interactively:
.br
      forth
.LP
To start the forth system, and compile a stand-alone diagnostic
from the file "test.f", saving a stand-alone bootable file in stand.test,
and copying the result to a server called venus, from where it may
be booted over the ethernet:
.br
      forth stand.exe test.f -s' " stand.test" save-stand'
.br
      rcp stand.test venus:/pub/stand
.SH FILES
files.exe	forth binary image
.br
/usr/local/lib/forth/files.exe
.br
       The usual forth image for Unix
.br
/usr/local/lib/forth/stand.exe
.br
       The image for compiling stand-alone programs.
.br
/usr/local/forth
.br
       The dynamic loader
.br
pi:/usr/local/lib/forth/*
.br
       Library packages
.SH "SEE ALSO"
FORTH Tutorial (in forth/doc/tutorial.doc)  
.LP
Roll Your Own (Hardware Diagnostics, That Is)  (in forth/doc/intro.doc)
.LP
Mastering Forth, by Martin Tracy and Anita Anderson
.LP
Starting Forth, by Leo Brodie
.LP
net.lang.forth on the Usenet is read by many knowledgeable Forth
enthusiasts, so it is a good source of information.
.LP
forth/doc contains many documentation files describing
various aspect of the Sun Forth system.
.SH DIAGNOSTICS
If the forth interpreter encounters a word that hasn't
been defined yet, it will print the name of the word
followed by a question mark and the name of the last
word that was compiled.  The compilation will then proceed,
so you will get a fairly-complete list of all such errors.
.SH BUGS
Sometimes the system will core dump after it has found a whole lot
of of undefined words.  This isn't too bad, since you have to fix
the undefined problems anyway before your program will work, but it
would be nicer if it didn't core dump.
.LP
It is possible to crash the Forth system from the keyboard,
causing a core dump.  This is because the system does not try
very hard to protect you from yourself.  One easy way to crash the
system is to store random stuff into addresses below HERE.n
This does not cause Unix to crash; it usually just causes the
Forth process to go away. It is usually possible to restart a
crashed stand-alone forth system from the ROM monitor by typing
>g10004
.LP
The documentation should be better organized.  It is currently
somewhat fragmented, in keeping with the Unix tradition.
