\ Test for in-line code forth compiler.

\ Word classes:
\ ICode - like code but when in compile state the code will be
\    copied to the dictionary.
\ Code  - defined as machine code.  When compiled, the dictionary
\    gets a  "bsr <code-address>"
\ Colon - defined as a series of forth words.  When compiled, the dictionary
\    gets a  "bsr <code-address>"
\
\ The view field distinguishes between these.
\ Icode - view field contains the length of the code sequence
\ Code, colon - view field contains 0
\
\ Virtual machine implementation:
\ sp - a6
\ rp - a7

0 warning !
: empty forth definitions [ last @ n>link 2- ] literal (forget) ;
create sn 100 allot  ( stack space for new scheme )
variable save-rp
variable tn  ( top of new stack )
sn 100 + constant ntop
: init-n ntop tn ! ;
init-n
: .sn
  ntop
  ntop tn @ - 4 / 0 ?do 4 - dup 2@ d. loop drop
;
: >n  ( d -- )  -4 tn +!  tn @ 2! ;
: n>  ( -- d )  tn @ 2@  4 tn +! ;
( execute a word defined using the new scheme )
code nexecute ( cfa -- )
   word save-rp #  d7 move  d7 a0 lmove
   rp a0 ) lmove   ( save the old rp )
   tn # d7 move  d7 a0  lmove   ( set the rp to the new stack area )
   a0 ) d7 move  d7 rp  lmove
   sp )+ d7 move 2 d7 addq d7 a0 lmove  long a0 ) jsr word
   tn # d7 move  d7 a0  lmove
   rp a0 ) move
   save-rp #  d7 move  d7 a0 lmove
   a0 ) rp lmove   ( restore the old rp )
c;
assembler definitions
   th 0f constant rp
   th 0e constant sp
   : next rts ;
only forth also definitions
vocabulary new
: do ( --namex )
  context @ >r new '  r> context !
  nexecute
;
new definitions

: constant ( l -- )
  create
  [ assembler ] long l# sp -) move next [ new ]
;
: variable ( -- )
  create
  [ assembler ] th 41fa , 6 , ( 6 pc d( a0 lea ) a0 sp -) move  next [ new ]
\  [ assembler ] here rel 8. d+ l# sp -) move rts [ new ]
  0 , 0 ,
; 
: uservar ( -- )
  create 4 ualloc
  [ assembler ] up d) a0 lea  a0 sp -) move  next [ new ]
;

\ Code for does> children:   6 pc d) a0 lea   a0 sp -) move  <does-clause> jmp

: n; state off ; immediate
: compile-call ( cfa -- )
( should check for out of 15-bit range, and if so, compile a jsr )
   >body ( cfa ) [ assembler ] bsr [ new ]
\  >body rel [ assembler ] l#) jsr [ new ]
;
: handle-cfa ( cfa f -- ) ( f is >0 if immediate )
  0>  if  execute
  else  compile-call
  then
;
: handle-word ( str -- f ) ( f is true if success )
  find ( cfa f | str 0 )
  dup
  if   handle-cfa    state @
  else drop count type ." ?" cr   false
  then
;
: n: ( -- )
  create  state on
  begin  bl word dup count nip
    if   handle-word  ( success-flag )
    else query  true
    then
  0= until
  [ assembler ] next [ new ]
;
code 1-l  ( -- 1 )
  long 1. l# sp -) move
  rts
c;

\ Literal numbers are compiled as   nnnnnnnn l# sp -) move

code execute   (s cfa -- )
     rts
c;

\ branches are compiled in-line as   target Bcc
\ how do loops work?

( label dofalse ) here  assembler  long  sp -) clr   rts
new
hex
code digit ( char base --- digit true | char afalse )
 long
   sp )+ d0 move ( base )   sp ) d1 move ( char )
 byte
         30 # d1 sub   ( dofalse ) dup bmi
         0a # d1 cmp
   0>= if
         31 # d1 cmp   ( dofalse ) dup bmi
         27 # d1 sub
   then
   d0 d1 cmp            ( dofalse ) dup bpl
 long
   d1 sp ) move
   -1. l# sp -) move
   rts
c;
drop ( dofalse )

decimal
( label done ) here
    assembler   long  a0 sp -) move  d1 sp -) move  rts   new
code skip  (s addr1 len1 char -- addr2 len2 )
 long
  sp )+ d0 move   sp )+ d1 move  1 d1 addq
  sp )+ a0 move
  begin  1 d1 subq  0<>
  while
    byte  a0 ) d0 cmp  ( done ) dup bne  word
    1 a0 addq
  repeat
  ( done ) dup bra
c;
code scan  (s addr1 len1 char -- addr2 len2 )
 long
  sp )+ d0 move  sp )+ d1 move  1 d1 addq
  sp )+ a0 move
  begin
    1 d1 subq  0<>
  while
    byte  a0 ) d0 cmp  ( done ) dup beq  word
    1 a0 addq
  repeat  ( done ) dup bra
c;
drop ( done )
code hash  (s str-addr voc-ptr -- thread )
 long
  sp )+ d1 move  sp )+ a0 move
  1 a0 addq   byte a0 )+ d0 move long
  ( for 4 threads ) 3. l# d0 and  d0 d0 add  d0 d1 add
  d1 sp -) move
  rts
c;
 
\ code enclose   (s addr  delim -- addr  n1 n2 n3 )
\  long
\    sp )+ d0 move   sp ) a0 move   a0 d6 lmove
\    begin  byte   a0 )+ d1 move   d1 d0 cmp  long  0<>
\    until   1 a0 subq
\    long   a0 d2 move   d6 d2 sub    d2 sp -) move
\    byte   a0 ) tst   long   0=
\    if   1 d0 addq   d0 sp -) move   1 d0 subq   d0 sp -) move  rts   then
\    begin  byte   a0 )+ d1 move   long   0=
\      if   d6 a0 suba   1 a0 subq   a0 sp -) move  a0 sp -) move  rts then
\      byte   d0 d1 cmp   long   0=
\    until  d6 a0 suba   1 a0 subq  a0 sp -) move
\                        1 a0 addq  a0 sp -) move
\    rts
\ c;
code cmove ( from_addr to_addr cnt --- )
     long  sp )+ d0 move
           sp )+ a1 move
           sp )+ a0 move
     here 4 + bra
     byte  a0 )+ a1 )+ move  long
     here 2- d0 dbra
     rts
c;
code cmove> ( from_addr to_addr cnt --- )
     long  sp )+ d0 move
           sp )+ a1 move
           sp )+ a0 move
     d0 a0 adda    d0 a1 adda
     here 4 + bra
     byte  a0 -) a1 -) move  long
     here 2- d0 dbra
     rts
c;
code um*  (s d1 d2 -- d )   ( doesn't really do 32-bit arithmetic )
long   sp )+ d0 move   sp )+ d1 move   d1 d0 mulu
   long   d0 sp -) move   rts
c;

code /mod (s d1 d1 -- Remainder Quotient )
   long   sp )+ d0 move
   long   sp )+ d1 move   d0 d1 divu
   d0 clr   word  d1 swap  d1 d0 move  long d0 sp -) move
            word  d1 swap  d1 d0 move  long d0 sp -) move
   rts
c;
assembler long new 
code and sp )+ d0 move  d0 sp ) and  rts  c;
code or  sp )+ d0 move  d0 sp ) or   rts  c;
code xor sp )+ d0 move  d0 sp ) eor  rts  c;
code not sp ) notw rts c;

code +   sp )+ d0 move  d0 sp ) add  rts  c;
code -   sp )+ d0 move  d0 sp ) sub  rts  c;
code negate    sp ) neg   rts  c;
code ?negate   sp )+ tst  0< if  sp ) neg  then   rts c;
code abs   sp )  tst  0< if sp ) neg  then   rts c;

code min   long
   sp )+ d0 move   sp ) d0 cmp
   < if   d0 sp ) move   then   next c;
code max   long
   sp )+ d0 move   sp ) d0 cmp
   > if   d0 sp ) move   then   next c;

code up@   up sp -) move   next  c;
code sp@   sp sp -) move   next  c;
code rp@   rp sp -) move   next  c;
code up! ( new-up --- ) sp )+ up move next c;
code sp!   sp )+ sp move   next c;   ( f83 )
code rp!   sp )+ rp move   next c;   ( f83 )
code >r    sp )+ rp -) move   next  c;
code r>    rp )+ sp -) move   next  c;
code r@    rp )  sp -) move   next  c;

code exit  rp )+ d0 move   next  c;  ( XXX may change for icode )
code unnest   rp )+ d0 move  next  c;   ( XXX may change for icode )
code tuck  (s n1 n2 -- n2 n1 n2 )
  sp )+ d0 move  sp )+ d1 move
  d0 sp -) move  d1 sp -) move  d0 sp -) move
  next 
c;
code nip (s n1 n2 -- n2 ) sp )+ sp ) move  next c;

assembler definitions
: leaveflag  ( condition -- )
( macro to assemble code to leave a flag on the stack )
( one of the operands of the comparison should be on the )
( top of the stack when the code executes.  It will be replaced )
( by the flag )
    long d0 [ forth ] swap [ assembler ] setif
    d0 extw  d0 extl  d0 sp -) move  next
;
new definitions
code 0=  sp )+ tst  0=  leaveflag c;
code 0<> sp )+ tst  0<> leaveflag c;
code 0<  sp )+ tst  0<  leaveflag c;
code 0<= sp )+ tst  0= if -1. l# sp )+ move next then
                   0<  leaveflag c;
code 0>  sp )+ tst  0= if sp )+ clr next then
                   0>= leaveflag c;
code 0>= sp )+ tst  0>= leaveflag c;
 
\ the odd sense of the comparisons is because the operands
\ come off the stack in the wrong order.
assembler definitions
: compare long sp )+ d0 move  sp )+ d0 cmp ;
new definitions

code <   compare >   leaveflag c;
code >   compare <   leaveflag c;
code =   compare 0=  leaveflag c;
code <>  compare 0<> leaveflag c;
code u>  compare u<  leaveflag c;
code u<= compare u>= leaveflag c;
code u<  compare u>  leaveflag c;
code u>= compare u<= leaveflag c;
code >=  compare <=  leaveflag c;
code <=  compare >=  leaveflag c;

code drop  sp )+ d0 move   next c;
code dup   sp ) sp -) move    next  c;
code over  4 sp d) sp -) move    next  c;
code swap  sp )+ d0 move  sp )+ d1 move
  d0 sp -) move  d1 sp -) move  next c;
code rot (s n2 n1 n0 -- n1 n0 n2 )
   sp )+ d1 move   sp )+ d2 move   sp ) d0 move
   d2 sp )  move   d1 sp -) move   d0 sp -) move   next c;
code -rot (s n2 n1 n0 -- n0 n2 n1 )
  sp )+ d2 move  sp )+ d0 move  sp ) d1 move
  d2 sp )  move  d1 sp -) move  d0 sp -) move  next
c;

code 2drop  (s d -- )     sp )+ d0 move  sp )+ d0 move   next c;
code 2dup   (s n1 n2 -- n1 n2 n1 n2 )
  4 sp d) sp -) move  4 sp d) sp -) move  next c;
code 2over  (s n4 n3 n2 n1 -- n4 n3 n2 n1 n4 n3 )
  12 sp d) sp -) move    12 sp d) sp -) move   next c;
code 2swap     (s d1 d2 -- d2 d1 )
  sp )+ d0 move   sp )+ d1 move   4 sp d) sp -) move  4 sp d) sp -) move
  d0 8 sp d) move   d1  12 sp d) move   next c;
code pick    (s nm ... n1 n0 k -- nm ... n2 n0 nk )
  sp )+ d0 move   d0 d0 add  d0 d0 add
  0 d0 sp di) sp -) move   next
c;
\ : roll    (s nk nk-1 ... n1 n0 k --- nk-1 ... n1 n0 nk )
\   >r  r@ pick   sp@ dup 2+
\   r> 1+ 2* 
\   cmove> drop
\ ;
code 1+   1 sp ) addq   next c;
code 2+   2 sp ) addq   next c;
code 1-   1 sp ) subq   next c;
code 2-   2 sp ) subq   next c;
code 2*   sp ) d0 move  1 # d0 asl  d0 sp ) move next c;
code 2/   sp ) d0 move  1 # d0 asr  d0 sp ) move next c;
code u2/  sp ) d0 move  1 # d0 lsl  d0 sp ) move next c;
code 4*   sp ) d0 move  2 # d0 asl  d0 sp ) move  next c;
code 8*   sp ) d0 move  3 # d0 asl  d0 sp ) move  next c;

code cset (s b addr -- )
  sp )+ a0 lmove   sp )+ d0 move
  byte  d0 a0 ) or  long  next c;
code creset (s b addr -- )
  sp )+ a0 move   sp )+ d0 move   d0 notw
  byte   d0 a0 ) and  long  next c;
code ctoggle (s b addr -- )
  sp )+ a0 lmove  sp )+ d0 move
  byte   d0 a0 ) eor  long  next c;
code on (s addr -- )
  sp )+ a0 lmove
  ( true ) -1 d0 moveq   d0 a0 ) move  next
c;
code off (s addr -- )
  sp )+ a0 lmove  a0 ) clr next c;

code +!  ( n addr -- )
  sp )+ a0 lmove     sp )+ d0 move     d0 a0 ) add
next c;

code toggle
  sp )+ d0 move   sp )+ a0 lmove
  byte   d0 a0 ) eor   long  next  c;

( requires alignment on a word boundary )
code @   sp ) a0 lmove   a0 ) sp ) move next c;
( doesn't require alignment on a word boundary )
code w@  sp ) a0 lmove   d0 clr
  byte   a0 )+ d0 move   word   8 # d0 lsl
  byte   a0 ) d0 move   long  d0 sp ) move    next  c;
code c@  sp ) a0 lmove   d0 clr
  byte   a0 ) d0 move   long  d0 sp ) move    next  c;
( requires alignment on a word boundary )
code !   sp )+ a0 lmove  sp )+ a0 ) move next c;
( doesn't require alignment on a word boundary )
code w!  sp )+ a0 lmove   sp )+ d0 move
  byte   d0  1 a0 d) move   word  8 # d0 lsr
  byte   d0 a0 ) move   long  next  c;
code c!  sp )+ a0 lmove   sp )+ d0 move
  byte   d0 a0 ) move   long  next c;
code 2@  sp )+ a0 lmove
  4 a0 d) sp -) move  a0 ) sp -) move  next c;
code 2!  sp )+ a0 lmove
  sp )+ a0 )+ move   sp )+ a0 )+ move   next c;

code fill ( start_addr cnt byte --- )
     sp )+ d1 move
     sp )+ a1 move
     sp )+ a0 move
     a0 a1 adda
     begin long a0 a1 cmpa u> while
        byte  d1 a0 )+ move   long
     repeat
     next
c;
code noop  next c;
