\ Some help for converting between C-style null-terminated strings
\ and forth style strings

\ Return the length of the C string
code cstrlen ( l.c-str -- len )
  sp )+ a0 lmove     a0 d0 lmove
  begin    a0 )+ byte tst  normal   0= until
  d0 a0 suba   1 a0 subq
  a0 sp -) move
c;
\ Same in high level:
\ : cstrlen ( l.c-str -- len)
\   dup   ( cstr cstr' )
\   begin dup c@l ( C strings may be anywhere in the machine's address space)
\   while 1+
\   repeat  ( cstr cstr' )
\   swap -
\ ;

\ Move from extended address space to the forth kernel address space
16\ code moveal ( l.from-addr to-addr cnt -- )
16\   sp )+ d0 move
16\   sp )+ a0 amove
16\   sp )+ a1 lmove
16\   here 4 + bra  ( jump to the dbra instruction )
16\      begin
16\        byte  a1 )+ a0 )+ move  normal
16\      d0 dbra
16\ c;

\ Leave the address and length for the argument C-style string
\ The funniness with the 16 bit system is because C strings nearly
\ always live outside of the address space of the Forth system,
\ so they have to be copied into the Forth system's address space.

: cscount ( l.c-string -- addr len )
  ldup cstrlen  ( l.c-string len)
16\  >r here r@ moveal here r>
;

\ Given a C-string, emplace it as a forth packed string at where
: fstr ( l.c-string where -- where)
  lnswap  cscount  ( where addr len)
  dup 255 > if ." C-string too long; truncating" cr drop 255 then
  >r  over 1+ r@ move  ( where)
  r>  over c!
;
\ Convert a forth-style packed string to a C-style null-terminated string
\ The forth string is first copied to here so appending the null won't
\ screw something up.
\ It might be nice to have a version of this that takes as an argument
\ the place to put the c-string, so that several c-strings could be passed
\ to a C routine.
: cstr ( forth_string --- c_string )
  count >r here r@ cmove
  0 here r> + c!   here n->a
;
