\ Emacs-style line editing for command lines.  Needs to be more
\ efficient at updating the screen, rather than doing everything
\ a character at a time.  See "install-line-editing" for functions
\ implemented and key bindings

\ To run this successfully under Unix, you have to
\ "stty nl cbreak -echo"

\ requires terminals.f
only forth also definitions

decimal

  ( maximum size of input buffer )
variable buflen
  ( address of start of input buffer )
variable bufstart
  ( number of characters currently in input buffer )
variable bufcurrent
  ( character that was most recently typed, or character to add )
variable lastchar
  ( position of cursor within line )
variable bufcursor   0 bufcursor !
  ( remembers size of previous line )
variable last-bufcurrent  0 last-bufcurrent !

only forth also definitions
: forward-character
  bufcursor @  bufcurrent @ <
  if   1  bufcursor +!  cursor-forward
  then
;
: backward-character
  bufcursor @
  if   -1 bufcursor +!
       cursor-backward
  then
;
: erase-next-character
  bufcursor @  bufcurrent @ <
  if  -1  bufcurrent +!
       bufstart @  bufcursor @   +   dup 1+  swap
       bufcurrent @  bufcursor @  -  ( # of chars to move )
       cmove
       delete-char
  then
;
: erase-previous-character
  bufcursor @
  if   backward-character  erase-next-character
  then
;
: insert-character
  bufcurrent @  1+   buflen @  min   bufcurrent !
  bufstart @  bufcursor @   +   dup dup 1+  ( addr addr addr+1 )
  bufcurrent @  bufcursor @ - 1-  cmove>    ( addr )
  lastchar c@  swap  c!  (  )
  lastchar c@
  1  bufcursor +!
  bufcurrent @ bufcursor @ =   ( character at-end? )
  if   emit  else  insert-char  then
;
: beginning-of-line
  bufcursor @ 0
  ?do   backward-character  loop
;
: end-of-line
  bufcurrent @  bufcursor @  -  0
  ?do   forward-character   loop
;
: finish-line
  end-of-line
  bufcurrent @ last-bufcurrent !
  bufcurrent @ span !
  buflen @ bufcurrent !
  cr
  0 bufcursor !
;
: kill-to-end-of-line
  bufcursor @ bufcurrent !
  kill-line
;
: grab-last-line
  last-bufcurrent @ dup  bufcurrent ! bufcursor !
  bufstart @  bufcurrent @  type
;
: retype-line
  cr
  bufstart @  bufcurrent @  type
  bufcurrent @ bufcursor @  - 0
  ?do  cursor-backward loop
;
: erase-blanks-backward ( -- )
  bufcursor @
  if
    bufstart @ bufcursor @ 1- bounds swap   ( bufstart bufend )
    do  i c@
         bl <> if leave then
         erase-previous-character
    -1 +loop
  then
;
: erase-non-blanks-backward ( -- )
  bufcursor @
  if
     bufstart @ bufcursor @ 1- bounds swap   ( bufstart bufend )
     do  i c@
         bl = if leave then
         erase-previous-character
     -1 +loop
   then
;
: erase-previous-word 
  erase-blanks-backward
  erase-non-blanks-backward
;
: erase-blanks-forward ( -- )
  bufcurrent @ bufcursor  @
  ?do  bufstart @ bufcursor @ + c@
       bl <> if leave then
       erase-next-character
  loop
;
: erase-non-blanks-forward ( -- )
  bufcurrent @     bufcursor  @
  ?do  bufstart @ bufcursor @ +  c@
      bl = if leave then
      erase-next-character
  loop
;
: erase-next-word 
  erase-blanks-forward
  erase-non-blanks-forward
;
: beginning-of-word
  bufcursor @
  if
    bufstart @ bufcursor @ 1- bounds swap   ( bufstart bufend )
    do  i c@
         bl = if leave then
         backward-character
    -1 +loop
  then
;
: end-of-word ( -- )
  bufcurrent @ bufstart @ +
  bufcursor  @ bufstart @ +
  ?do i c@
      bl = if leave then
      forward-character
  loop
;
: forward-word
  forward-character
  end-of-word
;
: backward-word
  backward-character
  beginning-of-word
;
: quote-next-character
  key lastchar c!  insert-character
;

only forth also definitions
vocabulary keys-forth

\ This really should be a token (tuser) variable, but it's not
tuser keys  ' keys-forth keys !

create name-buf 32 allot

: add-char-to-string ( str char -- )
  over ( str char str )
  count dup >r ( str char addr len )
  + c!  ( str )
  r> 1+ swap c!
;
: add-char-to-name ( str char -- )
  dup bl <
  if   ( control character so translate to ^ form )  ( str char )
     over ascii ^ add-char-to-string  ( str char )
     ascii A 1- +  ( str char' )
  then
  add-char-to-string
;
defer not-found
' beep is not-found

: do-command ( -- )
  name-buf lastchar c@  add-char-to-name
  name-buf keys @ >body >user vfind ( cfa true | addr false )
  if  execute else not-found then
;

defer printable-char
' insert-character is printable-char
: clear-name-buf ( -- )
  name-buf width @ blank  0 name-buf w!
;
: (newexpect   (s adr len -- )
   dup span ! buflen !  bufstart !
   0 bufcurrent !  0 bufcursor !
   begin   bufcurrent @  buflen @ < 
   while   key  dup  lastchar c!  dup bl <
     if    drop clear-name-buf  do-command
     else  127 = if  erase-previous-character  else  printable-char   then
     then
   repeat
;

only forth also keys-forth definitions

: ^F forward-character ;
: ^B backward-character ;
: ^A beginning-of-line ;
: ^E end-of-line ;
: ^D erase-next-character ;
: ^H erase-previous-character ;
: ^K kill-to-end-of-line ;
: ^M finish-line ;
: ^J finish-line ;
: ^Q quote-next-character ;
: ^L retype-line ;
: ^[ key lastchar c! do-command ;
: ^[h erase-previous-word ;
: ^[d erase-next-word ;
: ^[f forward-word ;
: ^[b backward-word ;
: ^[= grab-last-line ;

only forth also definitions
: emacs-edit
  ['] keys-forth keys !
  ['] (newexpect ['] expect (is
  ['] crlf    ['] cr     (is
;
: (cold-hook  (cold-hook  emacs-edit ;
