create exceptions.f

\ The number left on the stack by an exception name is actually
\ the address of the head of a linked list of exception frames
\ for that exception.  An exception frame is several items on the
\ return stack:
\     a link to the next frame,
\     a pointer back to the list head,
\     saved sp,
\     saved sp.
\ Each time that catch? executes, it creates a new frame, which is equivalent
\ to installing an exception handler.
\ The last frame in the list has a 0 link.

\ In addition to the linked list of frames, which correspond to handlers
\ for a particular exception, the exception list heads are themselves
\ linked together in a list.
\ The list head has two fields which are kept in the parameter field of
\ of the exception word:
\     pointer to the first frame for this exception
\     link to the next exception
\ The list head for the linked list of exceptions is the variable exceptions.

variable exceptions 0 exceptions !

: exception: ( --name )
  create   ( pf:  frames-list-pointer  next-exception-link )
  0 ,                                ( initially empty list of frames )
  here exceptions @ ,  exceptions !  ( link into exceptions list )
;

\ pop-exception removes an exception frame from the return stack
\ when the word which installed the frame exits
create pop-exception ( -- )
  ] r> ( old-exception-handler ) r> ( old exception-variable )  !
  r> drop r> drop
;

\ A word may explicitly remove its exception frame in order to propagate
\ the exception to the next higher handler.
: uncatch ( exception -- )
  dup @ ( exception handler )
  ?dup if @ swap !  then
;

\ When an exception is raised, but no handler is there to catch it,
\ uncaught gets executed.
create uncaught  ( exception -- )
  ] ." Exception " body> >name .id ." raised"
  \ Clear out exception handlers
  exceptions  \ head of linked list of exceptions
  begin  @  ?dup  while   0 over !   repeat
  (quit
;

\ Catch? installs an exception handler.
\ It returns false when first called, true if raised.
\ The handler code is put inside an "if  ...  then" or "if  ...  else ... then"
\ construct following catch?
code catch? ( exception -- f)
  sp )+ a0 move   \ exception list head in a0
  ip rp -) move   \ save ip
  sp rp -) move   \ save sp
  a0 rp -) move   \ save list head
  a0 ) rp -) move \ save previous frame address
  rp a0 )  move   \ make this frame the current handler
  pop-exception n# rp -) rmove   \ install code to remove this frame
  false l# sp -) move           \ and return false
c;

\ Raise causes the most-recently-installed exception handler for the
\ given exception to be invoked.
code raise ( exception -- )
  sp )+ a0 move       \ list head in a0
  a0 )  rp  cmpa  u<  \ are any handlers installed for this exception?
  if                  \ there is a handler
      a0 ) rp move    \ find the handler frame
   \ Unwind the other exceptions
      exceptions l#) a1 move
      begin 0<> while ( exception-link in a1 )
         /l a1 subq  ( point at exception pfa)
	    begin
	      a1 ) a2 move   0<> while ( exception-frame-addr in a2)
 	      a2 rp cmpa      u>  while
	      a2 ) a1 ) move         ( update this exception pointer)
 	    repeat
         /l a1 addq  ( point at exception-link)
         a1 ) a1 move
      repeat
\      rp )+ a0 ) move          \ Make previous handler current
\      rp )+ d0 move ( eat one) \ Don't need the handle
      /n 2 * rp d) sp move            \ Restore the stack pointer
      /n 3 * rp d) ip move            \ Restore the ip
      pop-exception n# rp -) rmove   \ re-install code to remove this frame
      true l# sp -) move       \ and return true
  else                    \ no handler installed
      a0 sp -) move       \ Push the exception handle
      uncaught n# ip rmove \ and execute the uncaught exception handler
  then
c;

exception: abort-exception
