\ structures with bit fields
: shift << ;
variable bitoffset
: bitalign ( offset size -- offset' size )
( moves to next byte boundary if the bit offset is nonzero )
  bitoffset @ if
    0 bitoffset !   swap 1+ swap
  then
;
: struct 0 bitoffset ! 0 ;
: field ( offset size -- newoffset )
    create bitalign over , +
    does> ( structure-base -- structure-member-address )
          @ +
;
\ reserves an n-bits field
: resbits ( offset #bits -- newoffset )
  bitoffset @ +   8 /mod  ( offset bitoffset_mod_8 bitoffset/8 )
  swap bitoffset !  +
;  
\ bits are allocated from left-right, or from msb to lsb.
\ however, they are numbered from lsb=0 to msb=7
\ left-right generates the correct bit number
: left-right-bit# ( #bits -- #bits bit# )
  bitoffset @ over +  8 swap -
  dup 0< abort" Bit fields can't cross byte boundaries"
;

\ structure describing the layout of information in the parameter field
\ of a word created by "bits"

struct ( bit-record )
  2 field bit-offset
  2 field bit#
  2 field #bits
constant bit-record-size

\ defines an n-bit field
: bits ( offset #bits --namex newoffset )
  create  over , left-right-bit# , dup ,  ( offset #bits )
  resbits
  does> ( struct_addr pfa -- byte_addr bit# #bits )
     >r
     r@ bit-offset @ +  ( byte_addr )
     r@ bit#       @    ( byte_addr bit# )
     r> #bits      @    ( byte_addr bit# #bits )
;

hex
create masks
  0 c, 1 c, 3 c, 7 c, 0f c, 1f c, 3f c, 7f c, ff c,
decimal

: bitmask ( n -- mask ) masks + c@ ;

: bit@ ( byte_addr bit# #bits -- n )
  rot c@           ( bit# #bits byte )
  rot negate shift ( #bits shifted-byte )
  swap bitmask and
;
: bit! ( n byte_addr bit# #bits -- )
  bitmask         ( n byte_addr bit# mask )
  rot >r          ( n bit# mask )
  rot over and    ( bit# mask masked-n )
  2 pick shift    ( bit# mask shifted-masked-n )
  -rot            ( shifted-masked-n bit# mask )
  swap shift      ( shifted-masked-n shifted-mask )
  not             ( shifted-masked-n shifted-inverted-mask )
  r@              ( shifted-masked-n shifted-inverted-mask byte_addr )
  c@ and          ( shifted-masked-n masked-old-value )
  or              ( new-value )
  r>  c!
;

\ 68000 assembly language version of bit@ and bit!

\ code bit@ ( byte_addr bit# #bits -- n )
\   word
\   sp )+ d0 move  ( d0 is #bits )
\   masks # d7 move   d7 a0 lmove  ( relocated address of mask table )
\   byte
\   0  d0  a0  di)  d0 move        ( mask to d0 )
\   word
\   sp )+ d1 move  ( d1 is bit# )
\   sp )+ d7 move   d7 a0 lmove   ( relocate 16 bit address to 32 bits )
\   long d2 clr
\   byte a0 ) d2 move
\   d1 d2 lsr
\   d0 d2 andw
\   word d2 sp -) move
\ c;

\ code bit! ( n byte_addr bit# #bits -- )
\   word
\   sp )+ d0 move  ( d0 is #bits )
\   masks # d7 move   d7 a0 lmove  ( relocated address of mask table )
\   byte
\   0  d0  a0  di)  d0 move        ( mask to d0 )
\   word
\   sp )+ d1 move  ( d1 is bit# )
\   sp )+ d7 move   d7 a0 lmove   ( relocate 16 bit address to 32 bits )
\   sp )+ d2 move  ( d2 is n )
\   d0 d2 andw
\   d1 d2 lsl ( masked, shifted, n in d2 )
\   d1 d0 lsl ( shifted mask in d0 )
\   d0 not ( inverted mask in d0 )
\   byte d0 a0 ) andw
\        d2 a0 ) orw
\   word
\ c;
