\ Forth Floating point package.
\
\ Implements the Forth Vendors Group proposed Forth Floating Point Standard.
\ Runs only under Unix, using the C floating point library to do most of
\ the work.
\ Has a separate floating point stack.
\ All floating point numbers are IEEE double-precision format.
\ Requires a C support library, which is usually kept in
\ /usr/local/lib/forth/float.{c,o}
\
\ Problems:
\ Floating point numeric input is not integrated into the forth interpreter.
\ Slower than it has to be, due to the overhead of calling the C library
\  routines.  If you want to use this for heavy numerical work, it needs
\  some performance work.
\ Numeric output does not strictly follow the Forth Floating Point Standard.
\  In particular the PLACES variable isn't implemented.

only forth definitions vocabulary floating
only forth also floating also definitions

requires clink.f
"" /usr/local/lib/forth/float.o clink

create float.f

8 constant /f
/f constant f#bytes

th 20 /f * constant /fstack
create fstack /fstack allot

/l ualloc user fpsave
/l ualloc user fp0  fstack /fstack + fp0 l!

: fp! fpsave l! ;
: fp@ fpsave l@ ;

: initf
  fp0 l@ fp!
;
initf
: fdepth
  fp0 l@   fp@ - /f /
;

: fp [ assembler ] a1 ;
: getfp   [ assembler ] ['] fpsave >body @ up d)  fp lmove ;
: putfp   [ assembler ] fp ['] fpsave >body @ up d) lmove  ;

\ assembler macro
: syscallcode ( -- )
  [ assembler ]
  w )  a0  lmove
  a0 )  jsr
;
: binaryfop: ( cfa -- )
  create  >body l@ l,
  ;code    getfp
  /l fp d) sp -) lmove
  fp ) sp -) lmove
  /f fp addq
  fp sp -) lmove
  putfp
  syscallcode
  word /f /l + #  sp adda normal
c;
' _fadd binaryfop: f+
' _fsub binaryfop: f-
' _fmul binaryfop: f*
' _fdiv binaryfop: f/

: binary2fop:
  create  >body l@ l,
  ;code    getfp
  /l fp d) sp -) lmove
  fp ) sp -) lmove
  /f fp addq
  /l fp d) sp -) lmove
  fp ) sp -) lmove
  putfp
  syscallcode

  word /f 2* #  sp adda normal
  d1 /l fp d) lmove
  d0 fp ) lmove
c;
' _pow binary2fop: f**

: unaryfop:
  create >body l@ l,
  ;code
  getfp
  /l fp d) sp -) lmove
  fp ) sp -) lmove
  syscallcode
  /f sp addq
  d1 /l fp d) lmove
  d0 fp ) lmove
c;

' _cos unaryfop: fcos
' _sin unaryfop: fsin
' _tan unaryfop: ftan
' _exp unaryfop: faln
' _log unaryfop: fln
' _sqrt unaryfop: fsqrt
' _log10 unaryfop: flog
\ ?? falog
' _atan unaryfop: fatan
' _asin unaryfop: fasin
' _acos unaryfop: facos
' _sinh unaryfop: fsinh
' _cosh unaryfop: fcosh
' _tanh unaryfop: ftanh

code float  ( l -- f )
  ' _ffloat >body l@  l#)  jsr
  /l sp addq
  getfp
  d1 fp -) lmove
  d0 fp -) lmove
  putfp
c;

code int
  getfp
  /l fp d) sp -) lmove
  fp ) sp -) lmove
  /f fp addq
  putfp
  ' _fint >body l@  l#)  jsr
  /f sp addq
  d0 sp -) lmove
c;
code fabs
  getfp  word th 7fff # fp ) andi
c;
code fnegate
  getfp  word th 8000 # fp ) eori
c;

code f! ( f addr -- )
  sp )+ a0 amove
  getfp
  fp )+ a0 )+ lmove
  fp )+ a0 )+ lmove
  putfp
c;
code f@ ( addr -- f )
  sp )+ a0 amove
  getfp
  /l a0 addq
  a0 ) fp -) lmove
  a0 -) fp -) lmove
  putfp
c;
: fvariable create /f allot ;
code fdrop
  getfp /f fp addq putfp
c;
code fswap
  getfp
  fp )+ d0 lmove
  fp )+ d1 lmove
  /l fp d) fp -) lmove
  /l fp d) fp -) lmove
  d1 /l 3 * fp d) lmove
  d0 /l 2 * fp d) lmove
c;

code fover
  getfp
  /f /l + fp d) fp -) lmove
  /f /l + fp d) fp -) lmove
  putfp
c;
code fdup
  getfp
  /l fp d) fp -) lmove
  /l fp d) fp -) lmove
  putfp
c;

code frot
  getfp
  /f /f + /l + fp d) d1 lmove
  /f /f +      fp d) d0 lmove
  /f /l + fp d)  /f /f + /l + fp d) lmove
  /f      fp d)  /f /f +      fp d) lmove
     /l   fp d)  /f      /l + fp d) lmove
          fp  )  /f           fp d) lmove
          d1             /l   fp d) lmove
          d0                  fp  ) lmove
c;

code fpop ( f -- l l )
  getfp
  /l fp d) sp -) lmove
  fp ) sp -) lmove
  /f fp addq
  putfp
c;

code fpush ( l l -- f )
  getfp
  /l sp d) fp -) lmove
  sp )     fp -) lmove
  /f sp addq
  putfp
c;

code f0=
  getfp
  fp )+ d0 lmove
  fp )+  d0 or
  0= if  d0 not  else  d0 clr then
  d0 sp -) lmove
  putfp
c;

code f0<
  getfp
  fp )+ d0 lmove
  d0 d1 lmove
  th 7fff.ffff l# d0 andw
  fp )+  d0 or
  0<>
  if  d1 tst 
     0< if -1. l# d0 lmove else  d0 clr then
  then  ( else d0 is already 0 )
  d0 sp -) lmove
  putfp
c;
: falog ( f -- f )  td 10 float fswap f** ;
\ : e.  ( f -- )
\   base @ >r
\   fdup f0= if ." .0E0 " fdrop exit then
\   fdup fabs
\   flog fdup             ( f log log )
\   fdup int float f-     ( f log log.frac )
\   falog                 ( f log mantissa )
\ \ could instead do a loop of multiply by 10 and print integer portion
\   td 10000000 float f*  ( f log scaled-mantissa )
\   frot                  ( log scaled-mantissa f )
\   f0< if ." -" then
\   int <# # # # # # # # ascii . hold # #> type
\            ( log )
\   ." E" int .
\   r> base !
\ ;
: >f  ( l -- fscaled )
  float
  dpl @ 0>
  if  td 10 float dpl @ float f** f/
  then
;
\ Convert a null-terminated cstr to a counted fstr
create fstrbuf td 128 allot
: fstr ( cstr-addr -- fstr-addr )
  dup
  td 127 0
  do dup  ( cstr cptr cptr )
     c@ 0= if  leave  then
     ca1+
  loop    ( cstr cstrend )
  over -  ( cstr clen )
  fstrbuf place
  fstrbuf
;
: f. ( f -- )
 fpop  _ftoaf lretval  ( cstr-addr )
 fstr count type space
 ldrop ldrop
;
: e. ( f -- )
 fpop  _ftoae lretval  ( cstr-addr )
 fstr count type space
 ldrop ldrop
;

: f< ( f1 f2 -- flag )
  f- f0<
;
: f> ( f1 f2 -- flag )
  fswap f- f0<
;
: f= ( f1 f2 -- flag )
  f- f0=
;
: f<= ( f1 f2 -- flag )
  f> 0=
;
: f>= ( f1 f2 -- flag )
  f< 0=
;
code (flit ( -- fp )
  getfp
   ip )+ fp -) lmove   ip )+ fp -) lmove
  putfp
c;
: fliteral ( fp -- )
  compile (flit
  fpop lswap l, l,
;   immediate

\ Kludge, kludge
: E ( l --number fscaled )
  state @
  if   \ If we're compiling, we have to grab the number from the code stream
    here /l - /token - token@
    ['] (llit =
    if    here /l - l@   /l /token + negate allot
    else ." E must be preceded by a number containing a decimal point"
         cr abort
    then
  then
  >f bl word number
  float td 10 float fswap f** f*
  state @  if  [compile] fliteral  then
; immediate
: fconstant ( fp -- )
  create fpop lswap l, l,
  ;code  getfp   w )+ fp -) lmove   w )+ fp -) lmove   putfp
c;

: f.s ( -- )
  fp0 l@
  fdepth 0 
  ?do   /f -   dup f@ f.    loop
  drop
;

: (cold-hook (cold-hook floating also initf ;
' (cold-hook is cold-hook
only forth floating also forth definitions
