\ For talking to a serial line which is presumed to connect to a modem.
\ Implements   M-KEY?   M-KEY   and  M-EMIT  which do the obvious things.
\ M-INIT must be called first to initialize the line.
\
\ Also implements M-CONNECT, which provides a "virtual terminal" connected
\ to that modem line.  To get out of virtual terminal mode, type ~.
\ (tilde period).

requires tasking.f
requires stty.f

hex
4004.667f lconstant FIONREAD

variable mfd   -1 mfd !
ascii ~ constant escchar
ascii . constant endchar

create old-modes  /sgttyb allot
: set-tty-modes ( -- )
  old-modes 0 gtty
  RAW 0 setmode    \ turn off input editing
;
: restore-tty-modes ( -- )
  old-modes 0 stty
;
\ Establish the baud rate and turn off input editing on the modem line
\ ANYP is 7 bits, accept any parity.
variable mlm  ANYP RAW or mlm !
: set-modem-line ( -- )
  td 1200 mfd @ setbaud
  mlm @ mfd @ setmode
;
\ Open and initialize the modem line
: m-init ( line-name -- )
  mfd @ -1 =
  if    2 swap u_open
        dup 0< abort" Can't open modem line"
        mfd !
  else drop
  then
  set-modem-line
;
variable mbuf
: m-key? ( -- flag )  \ Flag is true if there is a character waiting
  mbuf FIONREAD mfd @  u_ioctl  drop
  mbuf @ 0<>
;
: m-key  ( -- char )
  1 mbuf mfd @  u_read drop
  mbuf c@
;
: m-emit  ( char -- )
  mbuf c!
  1 mbuf mfd @ u_write drop
;
: m-key-pause ( char -- )
  begin  pause  m-key? until
  m-key
;
: key-pause
  begin  pause  key?   until
  key
;
\ Listen to the keyboard and send the characters to the modem
: >modem
  begin
     key-pause
     dup escchar =
     if  key-pause
         dup endchar =
	 if   true
	 else dup escchar <>  \ don't double ecsapes
              if escchar m-emit then
              m-emit  false
	 then
     else m-emit false
     then
  until
;
\ Listen to the modem and send the characters to the display
create xxbuf 80 allot
: modem>
  begin
     begin  pause  m-key? until
      80 xxbuf mfd @  u_read  ( nread )
      xxbuf swap  type
  again
;
task: m>-task
' modem> m>-task fork
: m-connect ( line-name -- )
  m-init
  set-tty-modes
  m>-task wake
  >modem
  m>-task sleep
  restore-tty-modes
;
\ Connect is the point of the whole exercise.  Connect sets up
\ a through connection to the modem so that stuff you type at the keyboard
\ goes out over the modem serial line and stuff that comes in over the
\ line is displayed on your screen.  To get back to forth, type ~. (tilde
\ period).
: connect ( -- )
  [""] /dev/ttyb m-connect
;
: m-type ( addr len -- )
  bounds ?do i c@ m-emit loop
;
\ Sends a string to the modem line and appends a newline
\ Used from Forth when not connected in transparent mode
: mcmd ( str -- )
  count m-type newline m-emit
;
