\ Path scan.  Takes a directory search path string of the form
\ directory1:directory2:directory3...
\ and converts it to an array of pointers to the several directory names.
\ A zero is stored into the last entry in the array.
\ The names are extracted from the string and stored in the dictionary
\ as packed strings.

\ Next-name extracts the first name from the string specified by addr len
\ A name is terminated either by a colon or by the end of the string
\ Addr' and len' are left pointing to the remaining portion of the
\ string, not including the colon

: next-name ( addr len -- addr' len' addr count )
  over >r
  begin
    dup if   over c@ ascii : <>
        else false
        then        ( addr' len' f )
  while
    1- swap 1+ swap
  repeat            ( addr' len' )
  over r@ -   ( addr' len' count )
  over if  rot 1+ -rot  swap 1- swap  then  ( Skip the colon)
  r> swap     ( addr' len' addr count )
;
  
( max-entries must be > 0 )
: path-scan ( str array max-entries -- )
  2dup /n * erase    ( addr len array max-entries )
  rot dup
  if  \ Only do this if the string is non-null
    count 2swap    ( addr len array max-entries )
    1- /n * bounds
    ?do   ( addr len )
        dup 0= if leave then
        next-name  ( addr' len' addr count )
        here i !
        dup 1+ allot
        i @ place
    /n +loop
    align
    if  
      1 abort" path-scan: Search path has more names than array"
    else drop  then
  then
;
