\ Terminal control.

only forth also definitions
create terminals.f

\ A terminal control array contains an entry for each terminal control
\ function.  Each entry consists of space for 3 cfas.
\ A terminal control function is executed by setting the ip to point
\ to the first cfa for that function.  The forth inner interpreter
\ them simply executes the string of cfa's.
\ All of the entries are initialized to be filled with the cfa for "unnest".
\ An entry that is not changed will thus behave as a no-op.
\ Any arbitrary forth code may be stuck in the space, so long as it doesn't
\ Need more than 2 cfa's worth of space (the third location is needed for
\ an "unnest").

\ Kludge: move 'word far away from here so word doesn't scribble on
\ the array we're building
1000 20 npatch 'word

\ Enough space for the common case, which is <ctl> <number> <exit>
/token 2 * /n + constant /termitem
variable terminal-type

code jumpto ( new-ip -- )
32\  sp )+ ip  move
16\  sp )+ ip  rmove
c;
\ High level version
\ : jumpto ( new-ip -- )
\   r> drop >r 
\ ;

\ creates a terminal control execution word
: termctl: ( offset -- new-offset )
  create  dup ,  /termitem +
  does> @ terminal-type @ + jumpto
;

0
termctl: #lines
termctl: #columns
termctl: beep
termctl: left
termctl: right
termctl: set-position
termctl: insert-char
termctl: delete-char
termctl: kill-line
termctl: kill-screen
termctl: insert-line
termctl: delete-line
termctl: erase-screen
termctl: save-position
termctl: restore-position
constant /termctl

: cursor-to-row-column ( row column-- )   2dup #out ! #line ! set-position  ;
: cursor-forward  ( -- )  right  1 #out +!  ;
: cursor-backward ( -- )  left  -1 #out +!  ;
: clear-screen ( -- )    erase-screen  0 #out !  0 #line !  ;
vocabulary terminals
terminals definitions

\ For save-cursor-position
variable saved-x
variable saved-y
variable saved-pause

td 27 constant escape

code getlit ( -- literal )
  rp )+ a0 move
  a0 )+ sp -) move
  a0 rp -) move
c;
: outlit ( -- ) getlit (emit ;
: ctl ( --arg )
   bl word 1+ c@ bl 1- and state @ 
   if    compile outlit , 
   else  (emit 
   then  
; immediate

: (esc ( character -- )  escape (emit getlit (emit ;
: esc ( --char )
   bl word 1+ c@ state @ 
   if    compile (esc , 
   else  escape (emit (emit 
   then  
; immediate

: (esc[ ( --arg ) escape (emit ascii [ (emit getlit (emit ;
: esc[  ( --arg )
   bl word 1+ c@ state @
   if    compile (esc[ ,
   else  escape (emit ascii [ ( 91 ) (emit (emit
   then
;
immediate

: for  ( --termctl-name )
   ' >body @ dup 0 /termctl between
      0= abort" `For' argument is not a terminal control word" 
   terminal-type @ + dp ! 
; immediate

: clear-termctl ( -- )  \ Init termctl to all no-ops
  /termctl 0
  do   ['] unnest   terminal-type @ i + token!
       /token
  +loop   ( source-addr )
;

: terminal:  ( --name )
  create
  here terminal-type !
  clear-termctl
  ]
  does> terminal-type !
;
: end-terminal ( -- )
  terminal-type @  /termctl +  dp !
  [compile] [
; immediate

: prototype ( --name )
  ' >body  terminal-type @  /termctl cmove
; immediate

: dumb-del  bl (emit ctl H  ;

terminal: dumb
for #lines           24
for #columns         80
for beep             ctl G
for left             ctl H
for insert-char      (emit
for delete-char      dumb-del
end-terminal

only forth also definitions
