requires case.f
requires terminals.f

create windows.f

variable last-window
nuser my-window
: struct 0 ;
: field ( offset size -- offset' )
  create
  over , +
  does>  @ my-window @ +
;
: int /n field ;

struct ( window )
  int row0
  int col0
  int #rows
  int #cols
  int w#line
  int w#out
constant /window

: window:  ( row0 col0  #rows #cols -- )
  create
   2swap
   swap , ,   swap , ,  0 ,  0 ,
;
23 0  10 80 window: console
console my-window !

: at ( row column -- )
  2dup #out !  #line !
  swap row0 @  + swap
       col0 @  +  set-position
;
: select-window  ( window -- )
  dup last-window @ <>
  if   last-window !
       #line @ #out @ at
  else drop
  then
;
: doreturn ( -- )
  #line @  0 at
;
: clear-line ( -- )
  \ If the right side of the window is at the right side of the screen,
  \ use the terminals kill line command
  col0 @ #cols @ +   #columns  =
  if    kill-line
  else  #cols @ #out @ - ( nchars )
        dup  0  ?do  bl (emit  loop
             0  ?do  bs (emit  loop
  then
;
: cursor-to-row-column ( row column-- )   2dup #out ! #line ! set-position  ;
: cursor-forward  ( -- )
  #out @ #rows @ <
  if right  1 #out +!  then
;
: cursor-backward ( -- )
  #out @ 0>
  if  left  -1 #out +!  then
;
: clear-screen ( -- )
  #rows @  0
  ?do
      i dup #line !  ( line ) 0 at  clear-line
  loop
  0 0 at
;
\ For Unix
\ create blbuf 150 allot  blbuf 150 bl fill
\ : clear-line
\   #cols @ #out @ -  blbuf 1 u_write drop
\   #line @ #out @ at
\ ;
\ Usually windows wrap around instead of scrolling, but a full width
\ window at the bottom of the screen will scroll.
\ Scrolling is only a good idea if the window is as wide as the screen
\ and the window is at the bottom.  If it's not full width, real scrolling
\ requires that the system keep an image on what's in the window.  If
\ the window isn't at the bottom, then the rest of the screen jumps
\ while you scroll the window.  This is a consequence of the decision
\ to run on any smart ascii terminal.  If the device model were a bitmap
\ or a character map instead of an ascii terminal, this would be different
: scroll ( -- )
  0 0 at
  col0 @ 0=   #cols @  #columns =  and   row0 @ #rows @ +  #lines =  and
  if
     delete-line   #rows @ 1-  0 at   insert-line
  then
;
: donewline ( -- )
  1 #line +!
  #line @  #rows @ >=
  if  scroll   else   linefeed (emit  then
  doreturn
  clear-line
;
: do-printable ( char -- )
  1 #out +!
  #out @  #cols @  >=
  if   donewline  1 #out +!   then
  (emit
;
variable tabstops  8 tabstops !
: dotab ( -- )
  #out @  tabstops @  mod  ( spaces into tabstop )
  tabstops @  swap -       ( spaces to move )
  spaces
;
: do-ctl ( char -- )
  case
      linefeed  of  donewline  endof
      carret    of  doreturn   endof
      control H of  #out @  if  -1 #out +! left  then  endof
      control I of  dotab   endof
  ( should handle form feeds too)
      ( default   ) ascii ^ do-printable
                    dup ascii @ + do-printable
  endcase
;
: wemit ( char -- )
  my-window @ select-window
  dup bl <
  if  do-ctl  else  do-printable  then
;
variable input-window
console input-window !
: pause-key? pause (key? ;
: pause-key
   begin
\     input-window @ select-window
     key?
   until
   (key
;
: install-windows
     erase-screen
     ['] (lf ['] cr (is
     ['] wemit  ['] emit  (is
     ['] w#out  ['] #out  (is
     ['] w#line ['] #line (is
;

\ should be in terminals
terminals also
: dark ( -- )    esc[ 7 ascii m (emit  ;
: light ( -- )   esc[ m ;
only forth also definitions

: (type ( addr len -- )
  bounds ?do i c@ (emit loop
;

: mode-line ( str -- )
  #line @  #out @     \ Remember the current position
  rot                 ( line out str )
  #rows @ 0 at
  dark
    count tuck ( len addr len )  (type  ( len )
    #cols @ swap ?do bl (emit loop
  light
  at     \ Put the cursor back where it was
;
