
: setbaud ; : set8bits ;

only forth also modem also definitions
decimal

variable modem-file   ( file descriptor for current file )
modem-file off
variable checksum
variable #errors    10 constant max#errors  variable #naks
variable receive-sector# variable send-sector#
variable expected-sector
 0 constant nul   24 constant can    1 constant soh
 4 constant eot    6 constant ack   21 constant nak
variable timer-init   variable timer
: long-timeout  8000 timer-init ! ;
: short-timeout 500 timer-init ! ;
short-timeout
: get-filename ( --name  str )
  bl word  dup count upper  pad 18 cmove  pad
;
: abort-end ( -- ) ( abort and clean up )
  ." aborting." cr  modem-file @ ?dup if  close  then quit
;
: ?interrupt ( -- ) ( aborts if user types control c )
  key? if  key control c =  if  abort-end   then  then
;
: init-file ( --name )  get-filename dup write open
  ( name fd ) dup modem-file ! ( name fd )
  if ." file exists.  clobber it? "
     key dup emit upc ascii y <>  if  abort-end then
  else    dup make 0=  abort" can't make file" ( name )
      write open  ?dup
      0= abort" can't open file"  modem-file !
  then
;
: receive-setup
  setbaud set8bits  ( initialize modem )
  init-file
  0 receive-sector# !   0 expected-sector !  0 #naks !
;
: timed-in ( -- char | -1 ) ( get a character unless timeout )
  timer-init @  timer !
  begin  m-key?  if  m-key  exit   then
         -1 timer +!   timer @
  0= until  -1
;
: gobble  ( -- ) ( eat character till they stop coming )
  begin  timed-in  -1 =
  until
;
: receive-error ( -- ) ( eat rest of packet and send a nak )
  gobble
  1 #naks +!  #naks @ max#errors >
  if  can m-emit   ." too many errors." cr
      abort-end
  then
  ." sent nak." cr  nak m-emit
;
create sector-buf  128 allot    variable sector-ptr
: save-sector  ( -- ) ( write out the sector )
  sector-buf 128   modem-file @ fputs
  carret emit  expected-sector @ .
  1 expected-sector +!
;

: normal-end ( -- ) ( clean up )
  ." reception completed" cr
  ack m-emit   modem-file @ close
;
: receive-header ( -- f ) ( true if header error )
  timed-in  dup  -1 =  if  exit then
  dup receive-sector# !
  timed-in  dup  -1 =  if  exit then
  255 xor <>
;
: .bogus-char ( char -- )
  base @ >r hex    cr dup 2 .r   r> base !
  ." h(" emit ." ) unexpectedly seen."
;

: receive-sector ( -- f ) ( true if runt sector )
  0 checksum !
  false   sector-buf
  128 0
  do   timed-in dup -1 =
       if  ( false sector-buf -1 )  rot drop swap leave then
       2dup  swap i +   c!   checksum +!
  loop ( false sector-buf ) drop
;
: receive-checksum ( -- f ) ( true if checksum error )
  timed-in dup -1 <>               ( char true  |  -1 false )
  if    checksum @ 255 and  <>  then
;


: receive-packet ( -- f ) ( true if end of transfer )
  false timed-in
    dup nul =  if drop exit then
    dup can =  if drop ." remote cancel" cr abort-end then
    dup eot =  if drop normal-end drop true exit then
    dup -1  =  if drop ." t-" receive-error exit then
    dup soh <> if .bogus-char receive-error exit then
  drop
  receive-header    if ." h-" receive-error exit then
  receive-sector    if ." s-" receive-error exit then
  receive-checksum  if ." c-" receive-error exit then
  save-sector   ack m-emit   0 #naks !
;


: receive ( -- )
  receive-setup
  gobble  ."  reception started" cr   nak emit
  begin  ?interrupt  receive-packet
  until
;









\ xodem                                               wmb30oct84
: wait-ack ( -- ) ( wait for ack or can )
  0 #errors !
  begin  #errors @  max#errors >  #naks @  max#errors > or
    if  ." too many errors." cr can m-emit  abort-end  then
    ?interrupt
    timed-in
    dup -1  = if drop  1 #errors +!  ." timeout"  cr       else
    dup can = if drop  ." remote cancel" cr abort-end      else
    dup ack = if drop  #naks off  exit                     else
    dup nak = if drop  1 #naks +! exit                     else
      .bogus-char
    then then then then
  again
;

: read-sector ( -- end-of-file? )
  sector-buf  128 control z fill
  sector-buf  128  modem-file @  fgets 0=
;
: send-header ( -- ) ( header is  soh sector#  sector#not )
  soh m-emit
  send-sector# @  255 and  dup m-emit
  255 xor m-emit
  carret emit  send-sector# @ .
;





: send-sector ( -- )
  0 checksum !
  sector-buf  128 0
  do   dup i + c@  dup m-emit    checksum +!
  loop drop
;
: send-checksum ( -- )
  checksum c@ m-emit
;
: end-send ( -- )
  modem-file @ fclose
  begin  eot m-emit  wait-ack   #naks @ 0=
  until
  ." file sent."
;

: init-send-file ( --name )
  get-filename  read  open
  dup 0= abort" can't open file"  modem-file !
;
: send-setup
  setbaud set8bits  init-send-file
  ." file size: " modem-file @ fsize l@  ul. cr
  0 send-sector# !   0 #naks !
;

: send ( --name )
  send-setup
  short-timeout gobble long-timeout
  wait-ack cr
  begin  ?interrupt
    #naks @ 0=
     if  read-sector  if  end-send  exit  then  then
     send-header
     send-sector
     send-checksum
     wait-ack
     #naks @ 0=  if  1 send-sector# +! then
  again
;

