/*	rechod.c

	rechod [-v]

	Server for recho, which is a
	program to bounce packets off a UDP echo server and measure
	the round-trip throughput.  Works through internet gateways.
	Based on Tony West's recho program (the source for rechod
	was lost).

	-v option causes it to keep statistics and print
	   on stderr when receiving special packet.

	David Goldberg
	Copyright: Sun Microsystems Inc. 1983

	Written:	11/16/83
*/

/* #define DEBUG */
/* #define FLIP			/* complement first byte */
#define streq(a,b) (strcmp(a,b) == 0)
#define DUMP_CODE 27
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <netdb.h>
#include <stdio.h>

#define NULL 0
#define ERROR -1
#define TRUE 1
#define FALSE 0

/* PMAX is the upper (exclusive) limit on packet size used in the test */
#define PMAX		2048

char buf[PMAX];
int stat[PMAX];
extern int errno;

main(argc, argv)
char **argv;
{
    struct sockaddr_in  server;
    struct sockaddr_in  from;
    struct hostent *h;
    int     s,
            fromlen;
    int     sz,
            vflag;

    register int    i;

    struct servent *sp;
    struct protoent *pp;

    if (argc == 2 && streq (argv[1], "-v"))
	vflag = 1;
    else
	vflag = 0;

/* spawn a subprocess and make the parent process go away */

    i = fork ();
    if (i == -1) {
	perror ("rechod: fork");
	exit (1);
    }
    if (i != 0)			/* parent exits */
	exit (0);

/* Lookup the echo service and the udp protocol in the services database */

    if ((sp = getservbyname ("echo", "udp")) == NULL) {
	printf ("Sorry, echo service not in services database\n");
	exit (1);
    }

    if ((pp = getprotobyname ("udp")) == NULL) {
	printf ("Sorry, UDP protocol not in protocols database\n");
	exit (1);
    }

/* Open a socket */

    if ((s = socket (AF_INET, SOCK_DGRAM, pp -> p_proto)) == 0) {
	printf ("Sorry, can't get a socket\n");
	exit (1);
    }


/* Bind the socket to the server */

    bzero ((char *) & server, sizeof (server));
    server.sin_addr.s_addr = INADDR_ANY;
    server.sin_family = AF_INET;
    server.sin_port = sp -> s_port;
    if (bind (s, &server, sizeof (server)) == -1) {
	perror ("rechod: bind");
	exit (1);
    }

/* Close all the file descriptors if the daemon is in silent mode */

    if (!vflag) {		/* if vflag on, keep terminal connection 
				*/
	for (i = 0; i < 10; i++)
	    if (i != s) (void) close (i);
	(void) open ("/", 0);
	(void) dup2 (0, 1);
	(void) dup2 (0, 2);
	i = open ("/dev/tty", 2);
	if (i >= 0) {
	    ioctl (i, TIOCNOTTY, 0);
	    (void) close (i);
	}
    }

/* Forever receive and echo packets */

    while (1) {

/* Receive a packet */
	fromlen = sizeof (from);
	if ((sz = recvfrom (s, buf, PMAX, 0, &from, &fromlen)) == -1) {
	    perror ("rechod: recvfrom");
	    continue;
	}

#ifdef FLIP
	buf[0] = ~buf[0];
#endif

/* Echo it back to the sender */
	if (sendto (s, buf, sz, 0, &from, fromlen) == -1)
	    perror ("rechod: sendto");

#ifdef DEBUG
/* Announce the reception of the packet */
	h = gethostbyaddr (&from.sin_addr, sizeof (from.sin_addr), AF_INET);
	if (h != 0)
	    printf ("got packet of size %d from %s\n", sz, h -> h_name);
#endif

/* If in verbose mode, respond to a dump-statistics request */
	if (vflag) {
	    if (sz == 1 && buf[0] == DUMP_CODE) {
		fprintf (stderr, "\necho server statistics\n");
		for (i = 0; i < PMAX; i++)
		    if (stat[i] != 0) {
			fprintf (stderr, "got %10d packets of size %10d\n",
				stat[i], i);
			stat[i] = 0;
		    }
	}
/* otherwise just accumulate statistics */
	else  /* histogram based on size of packet */
		if (sz >= 0 && sz < PMAX)
		    stat[sz]++;
	}
    }
}

