\ This program converts a binary file into a form suitable for downloading
\ to a Data-IO prom programmer.  The format used is Intel Hex Format,
\ which is select code 83 on the Data-IO.
\ 
\ To use:
\ % forth hexify.f -s "hexify your-binary-file-name"
\ %
\ On the Data-IO: 
\ <select> 83  <copy> <port> <ram> <start> <start>
\ % cat foo.hex >/dev/tty04  # or whatever device the Data-IO is connected to

hex

10 constant bytes/line
variable out-addr
create line-buf 64 allot
variable outfile
variable checksum
variable end-of-file?

: .2   (s u -- ) th ff and n->l <# # # #> 2dup upper type ;
: .4   (s u -- ) n->l <# # # # # #> 2dup upper type ;

: get-line
  bytes/line 0 do th ff line-buf i + c! loop
  bytes/line 0
  do
    cf @  fgetc
    dup  th -1 =
    if   drop end-of-file? on leave
    else line-buf i + c!
    then
  loop
;
: file-emit output-file top@ fputc ;
: restore-output
  output-file pop  flushbuf drop
  output-file pop  ['] emit (is
;
: change-output ( fd -- )
  ['] emit >is token@  output-file push
  ['] file-emit ['] emit (is
  output-file push
;
: put-line
  bytes/line  checksum !
  ." :"  bytes/line .2
  out-addr @ 8 >> th ff and checksum +!
  out-addr @ th ff and checksum +!
  out-addr @ .4
  ." 00"
  bytes/line  0  do
     line-buf i + c@
     dup checksum +!
     .2
  loop
  checksum @ ff and
  100 swap -  .2
  (lf
  bytes/line out-addr +!
;
: (hexify ( input-file  -- )
  cf !
  [""] foo.hex make drop
  [""] foo.hex write open
  outfile !
  outfile @ change-output
  end-of-file? off
  0 out-addr !
  begin
    get-line  put-line
    end-of-file? @
  until
  ." :0000000000" (lf
  restore-output
  cf @ close
  outfile @ close
;
: .instructions ( -- )
   ." On the Data-IO:" cr
   ." <select> 83  <copy> <port> <ram> <start> <start>" cr
   ." % cat foo.hex >/dev/tty04" cr
;
: hexify ( --input-file-name )
  hex
  bl word read open ?dup
  if   (hexify   .instructions
  else ." Can't open input file" cr
  then
;
