\ Writes out a stand-alone bootable Forth file in a.out format.
\ Expects these words to be already defined:
\ patchboot  ( -- )
\    Mucks about with the entry locations (the first few bytes of the image)
\    to do whatever is necessary for booting.
\ stand-cold-hook ( -- )
\    Any high level initialization that is appropriate.
\ stand-key? ( -- flag )    check for a character waiting
\ stand-key  ( -- char )    get a character
\ stand-emit ( char -- )    output a character

: stand-prompt
  ?interactive @  if .ok then
;
: stand-cold-hook
  stand-cold-hook  ( first execute the existing cold hook)
  ['] stand-key?   ['] key?   (is
  ['] stand-key    ['] key    (is
  ['] stand-emit   ['] (emit  (is
  ['] stand-prompt ['] prompt (is
  ['] noop         ['] where  (is
  up@ link !           ( initialize the link field for the main task)  

  only forth also definitions
  hex ?interactive on
;
hex
: 0., 0. l, ;
\ ( demand-paged format )
( old format )
create header 107. l, 0., 0., 0., 0., 4000. l, 0., 0.,
decimal

: makeheader ( set up an "a.out" style header )
  here user-size + origin - n->l   header 4 +   l! ( size )
;
: putheader  ( write out the header )
  td 32   header fd @  u_write drop
;
: user-init-token ( val cfa-of-user-var --- )
  >body @ pad + token!
;

: make-init-table ( --- )
  up@ pad user-size cmove  ( make a copy of the user variables at pad )
  \ Find the most recent stand-cold-hook.  May not necessarily be the
  \ one defined in this file.
  [""] stand-cold-hook find  ( cfa true | name false )
    0= abort" make-init-table can't find stand-cold-hook"
    ['] cold-hook user-init-token
;
  
: save-stand ( str --- )
  default-mode over u_creat fd ! ( make a new file )
  fd @ -1 = 
  if   ." can't create " ".
  else drop ( get rid of filename )
      makeheader  putheader
  ( put top address of user init table at location 8 )
      here user-size + origin 8 + token!
      patchboot
      here origin -   ( current-size )  origin
          fd @  u_write drop ( dictionary )
      make-init-table user-size pad fd @ u_write drop

      fd @ u_close
  then
;
decimal
